"""This submodule implements a registry of Utopia models, which is used to
provide the required model information to the frontend and use it throughout.
"""
from ..exceptions import BundleExistsError, ModelRegistryError
from .entry import ModelRegistryEntry

# Make names of class definitions available
from .info_bundle import ModelInfoBundle

# Import the registry class only as "private"; should only be instantiated once
from .registry import ModelRegistry as _ModelRegistry

MODELS = _ModelRegistry()
"""The model registry"""

# Make utility functions available which work on the created model registry
from .utils import get_info_bundle, load_model_cfg



"""This module provides the :py:class:`.GraphPlot` class."""

import copy
import logging
import os
import warnings
from typing import Any, Callable, Dict, Sequence, Tuple, Union

import matplotlib as mpl
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np

from ._mpl import ColorManager

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------

POSITIONING_MODELS_NETWORKX = {
    "spring": nx.spring_layout,
    "circular": nx.circular_layout,
    "shell": nx.shell_layout,
    "bipartite": nx.bipartite_layout,
    "kamada_kawai": nx.kamada_kawai_layout,
    "planar": nx.planar_layout,
    "random": nx.random_layout,
    "spectral": nx.spectral_layout,
    "spiral": nx.spiral_layout,
}
"""Available networkx node layout options"""

# -----------------------------------------------------------------------------


class GraphPlot:
    """This class provides an interface for visualizing a
    :py:class`networkx.Graph` object or a graph created from a
    :py:class:`~utopya.eval.groups.GraphGroup`.
    """

    def __init__(
        self,
        g: nx.Graph,
        *,
        fig: "matplotlib.figure.Figure" = None,
        ax: "matplotlib.axes.Axes" = None,
        select: dict = None,
        positions: dict = None,
        nodes: dict = None,
        edges: dict = None,
        node_labels: dict = None,
        edge_labels: dict = None,
        mark_nodes: dict = None,
        mark_edges: dict = None,
    ):
        """Initializes a ``GraphPlot``, which provides drawing utilities for a
        fixed graph.

        The drawing kwargs are stored and used when calling
        :py:meth:`.draw`.

        A ``GraphPlot`` can also be initialized from a
        :py:class:`~utopya.eval.groups.GraphGroup` via the
        :py:meth:`.from_group` classmethod.

        If drawing multiple times from the same ``GraphPlot`` instance, be
        aware that it only keeps track of the nodes/edges/labels/colorbars that
        were last associated with it. Use
        :py:meth:`.clear_plot` before
        re-drawing on the same axis.

        .. note::

            For some graph drawing kwargs it is possible to configure an
            automatized mapping from node/edge properties.
            This *property mapping* has the following syntax:

            .. code-block:: yaml

                some_layout_property:
                  from_property: my_node_or_edge_property
                  scale_to_interval: [low_lim, up_lim]

            The ``from_property`` specifies the node or edge property to be
            mapped from. If ``scale_to_interval`` is given, the layout property
            values are rescaled linearly the specified interval.

        Args:
            g (networkx.Graph): The networkx graph object
            fig (matplotlib.figure.Figure, optional): The matplotlib figure
                used for drawing
            ax (matplotlib.axes.Axes, optional): The matplotlib axes used for
                drawing
            select (dict, optional):
                Draw only a subgraph induced by a selection of nodes. Either
                select a list of nodes by passing the ``nodelist`` argument or
                do a radial node selection by specifying a ``center`` node and
                the ``radius``. The following arguments can be passed
                additionally:

                open_edges (bool, optional):
                    Whether to draw the edges for which only one of source and
                    destination is in the set of selected nodes. Disabled by
                    default.
                drop (bool, optional):
                    Whether to remove the non-selected nodes from the graph.
                    If False, *all* nodes are passed to the node positioning
                    model. Enabled by default.

            positions (dict, optional):
                Configuration for the node positioning. The following arguments
                are available:

                from_dict (dict, optional):
                    Node positions (2-tuples) keyed by node. If given, the
                    layouting algorithm given by the ``model`` argument will be
                    ignored.

                model (Union[str, Callable], optional):
                    The layout model that is used to calculate the node
                    positions (default=``spring``). Available
                    `networkx layout models <https://networkx.github.io/documentation/stable/reference/drawing.html#module-networkx.drawing.layout>`_
                    are: ``spring``, ``circular``, ``shell``, ``bipartite``,
                    ``kamada_kawai``, ``planar``, ``random``, ``spectral``,
                    ``spiral``.
                    See :py:data:`.POSITIONING_MODELS_NETWORKX`.

                    If installed, `GraphViz <https://pypi.org/project/graphviz/>`_
                    models can be selected with a prepended ``graphviz_``.
                    Options depend on the ``GraphViz`` version but may include:
                    ``dot``, ``neato``, ``fdp``, ``sfdp``, ``twopi``,
                    ``circo``. (Passed as ``prog`` argument to
                    :py:func:`~networkx.drawing.nx_agraph.graphviz_layout`).

                    If the argument is a callable, it is invoked with the graph
                    as the first positional argument and is expected to return
                    networkx-compatible node positions, i.e. a mapping from
                    nodes to a 2-tuple denoting the position.
                further kwargs:
                    Passed on to the chosen layout model.

            nodes (dict, optional):
                Drawing configuration for the nodes. The following arguments
                are available for property mapping:
                ``node_size``, ``node_color``, ``alpha``.
                The following arguments are allowed:

                node_size (scalar or sequence of scalars, optional):
                    The node size (default=300). Available for property
                    mapping. Can be mapped directly from the nodes' ``degree``,
                    ``in_degree``, or ``out_degree`` by setting the
                    ``from_property`` argument accordingly.
                node_color (color or sequene of colors, optional):
                    Single color (string or RGB(A) tuple or numeric value)
                    or sequence of colors (default: '#1f78b4'). If numeric
                    values are specified they will be mapped to colors using
                    the cmap and vmin, vmax parameters.

                    If mapped from property it may contain an additional
                    ``map_to_scalar``, which is a dict of numeric target values
                    keyed by property value. This allows to map from non-numeric
                    (e.g. categorical) properties.
                cmap (optional):
                    The colormap. Passed as ``cmap`` to
                    :py:class:`~utopya.eval.plots._mpl.ColorManager`.
                cmap_norm (optional):
                    The norm used for the color mapping. Passed as ``norm`` to
                    :py:class:`~utopya.eval.plots._mpl.ColorManager`.
                    Is overwritten, if a discrete colormap is specified in
                    ``cmap``.
                colorbar (dict, optional):
                    The node colorbar configuration. The following arguments
                    are allowed:

                    enabled (bool, optional):
                        Whether to plot a colorbar. Enabled by default if
                        ``node_color`` is mapped from property.
                    labels (dict, optional):
                        Colorbar tick-labels keyed by tick position (see
                        :py:meth:`~utopya.eval.plots._mpl.ColorManager.create_cbar`).
                    tick_params (dict, optional):
                        Colorbar axis tick parameters
                    label (str, optional):
                        The axis label for the colorbar
                    label_kwargs (dict, optional):
                        Further keyword arguments to adjust the aesthetics of
                        the colorbar label
                    further kwargs:
                        Passed on to :py:meth:`~utopya.eval.plots._mpl.ColorManager.create_cbar`.

                further kwargs:
                    Passed to `draw_networkx_nodes <https://networkx.github.io/documentation/stable/reference/generated/networkx.drawing.nx_pylab.draw_networkx_nodes.html>`_
                    when calling :py:meth:`.draw`.

            edges (dict, optional):
                Drawing configuration for the edges. The following arguments
                are available for property mapping: ``edge_color``, ``width``.

                The ``edge_color``, ``edge_cmap``, and ``colorbar`` arguments
                behave analogously for the edges as nodes.node_color,
                nodes.cmap, and nodes.colorbar for the nodes. Any further
                kwargs are (after applying property mapping), passed on to
                `draw_networkx_edges <https://networkx.github.io/documentation/stable/reference/generated/networkx.drawing.nx_pylab.draw_networkx_edges.html>`_
                when calling :py:meth:`.draw`.

                If arrows are to be drawn (i.e. for directed edges with
                arrows=True), only norms of type matplotlib.colors.Normalize
                are allowed.

            node_labels (dict, optional):
                Drawing configuration for the node labels. The following
                arguments are allowed:

                enabled (bool, optional):
                    Whether to draw node labels. Disabled by default. If
                    enabled, nodes are labeled by their index by default.
                show_only (list, optional):
                    If given, labels are drawn only for the nodes in this list.
                labels (dict, optional):
                    Custom text labels keyed by node. Available for property
                    mapping.
                format (str, optional):
                    If ``labels`` are mapped from property this format string
                    containing a ``label`` key is used for all node labels.
                decode (str, optional):
                    Decoding specifier which is applied to all property values
                    if ``format`` is used.
                further kwargs:
                    Passed on to `draw_networkx_labels <https://networkx.github.io/documentation/stable/reference/generated/networkx.drawing.nx_pylab.draw_networkx_labels.html>`_
                    when calling :py:meth:`.draw`.

            edge_labels (dict, optional):
                Drawing configuration for the edge labels. The following
                arguments are allowed:

                enabled (bool, optional):
                    Whether to draw edge labels. Disabled by default. If
                    enabled, edges are labeled by their (source, destination)
                    pair by default.
                show_only (list, optional):
                    If given, labels are drawn only for the edges (2-tuples)
                    in this list.
                edge_labels (dict, optional):
                    Custom text labels keyed by edge (2-tuple). Available for
                    property mapping.
                format (str, optional):
                    If ``edge_labels`` are mapped from property this format
                    string containing a ``label`` key is used for all edge
                    labels.
                decode (str, optional):
                    Decoding specifier which is applied to all property values
                    if ``format`` is used.
                further kwargs:
                    Passed on to `draw_networkx_edge_labels <https://networkx.github.io/documentation/stable/reference/generated/networkx.drawing.nx_pylab.draw_networkx_edge_labels.html>`_
                    when calling :py:meth:`.draw`.

            mark_nodes (dict, optional):
                Mark specific nodes by changing their edgecolor. Either specify
                a ``color`` for a list of nodes (``nodelist``), or specify a
                ``colors`` dictionary of colors keyed by node. Updates an
                existing ``nodes.edgecolors`` entry.

            mark_edges(dict, optional):
                Mark specific edges by changing their color. Either specify a
                ``color`` for a list of edges (``edgelist``), or specify a
                ``colors`` dictionary of colors keyed by edge (2-tuple).
                Updates an existing ``edges.edge_color`` entry.
        """
        # Set matplotlib figure and axis
        self.fig = fig if fig is not None else plt.gcf()
        self.ax = ax if ax is not None else self.fig.gca()

        self._g = g.copy()
        self._nodes_to_draw = None
        self._edges_to_draw = None
        self._nodes_to_shrink = None
        self.positions = None

        self._select_subgraph(**(select if select else {}))
        self.parse_positions(**(positions if positions else {}))

        # Drawing configurations
        # TODO With networkx v2.6 only FancyArrowPatches will be used for
        #      the edges (https://github.com/networkx/networkx/pull/4360).
        #      Then, there will be no simple way of showing edge colorbars.
        #      Remove the distinction between directed graphs (with arrows
        #      enabled) and undirected graphs at various places.
        self._node_colormanager = None
        self._edge_colormanager = None
        self._node_cbar_kwargs = {}
        self._edge_cbar_kwargs = {}
        self._show_node_cbar = False
        self._show_edge_cbar = False
        self._show_node_labels = False
        self._show_edge_labels = False
        self._node_kwargs = {}
        self._edge_kwargs = {}
        self._node_label_kwargs = {}
        self._edge_label_kwargs = {}

        self.parse_nodes(**(nodes if nodes else {}))
        self.parse_edges(**(edges if edges else {}))
        self.parse_node_labels(**(node_labels if node_labels else {}))
        self.parse_edge_labels(**(edge_labels if edge_labels else {}))
        self.mark_nodes(**(mark_nodes if mark_nodes else {}))
        self.mark_edges(**(mark_edges if mark_edges else {}))

        # matplotlib objects
        self._mpl_nodes = None
        self._mpl_edges = None
        self._mpl_node_labels = None
        self._mpl_edge_labels = None
        self._mpl_node_cbar = None
        self._mpl_edge_cbar = None

    # .........................................................................
    # Properties

    @property
    def g(self):
        """Get a deep copy of the graph associated with this instance.

        Returns:
            networkx.Graph: The graph object
        """
        return self._g.copy()

    # .........................................................................
    # Drawing methods

    def draw(
        self,
        *,
        fig: "matplotlib.figure.Figure" = None,
        ax: "matplotlib.axes.Axes" = None,
        positions: dict = None,
        nodes: dict = None,
        edges: dict = None,
        node_labels: dict = None,
        edge_labels: dict = None,
        mark_nodes: dict = None,
        mark_edges: dict = None,
        suppress_cbar: bool = False,
        update_colormapping: bool = True,
        **add_colorbars,
    ):
        """Draws the graph associated with the ``GraphPlot`` using the current
        drawing configuration.

        The current drawing configuration may be temporarily updated for this
        plot. The respective arguments accept the same input as in
        :py:meth:`.__init__`.

        Args:
            fig (matplotlib.figure.Figure, optional): matplotlib figure
            ax (matplotlib.axes.Axes, optional): matplotlib axis
            positions (dict, optional): Position configuration. If given,
                the current positions are replaced. If using a node positioning
                model the positions are recalculated.
            nodes (dict, optional): Temporarily updates the node-kwargs
            edges (dict, optional): Temporarily updates the edge-kwargs
            node_labels (dict, optional): Temporarily updates the
                node-label-kwargs
            edge_labels (dict, optional): Temporarily updates the
                edge-label-kwargs
            mark_nodes (dict, optional): Temporarily mark nodes kwargs
            mark_edges (dict, optional): Temporarily mark edges kwargs
            suppress_cbar (bool, optional): Whether to suppress the drawing of
                colorbars
            update_colormapping (bool, optional): Whether to reconfigure the
                nodes' and edges'
                :py:class:`~utopya.eval.plots._mpl.ColorManager`
                (default=True). If True, the respective configuration entries
                are ignored. Set to False if doing repetitive plotting with
                fixed colormapping.
            **add_colorbars: Passed to
                :py:meth:`.add_colorbars`
        """
        # Cache the current drawing configuration
        node_colormanager_cache = copy.deepcopy(self._node_colormanager)
        edge_colormanager_cache = copy.deepcopy(self._edge_colormanager)
        node_cbar_kwargs_cache = copy.deepcopy(self._node_cbar_kwargs)
        edge_cbar_kwargs_cache = copy.deepcopy(self._edge_cbar_kwargs)
        show_node_cbar_cache = self._show_node_cbar
        show_edge_cbar_cache = self._show_edge_cbar
        show_node_labels_cache = self._show_node_labels
        show_edge_labels_cache = self._show_edge_labels
        node_kwargs_cache = copy.deepcopy(self._node_kwargs)
        edge_kwargs_cache = copy.deepcopy(self._edge_kwargs)
        node_label_kwargs_cache = copy.deepcopy(self._node_label_kwargs)
        edge_label_kwargs_cache = copy.deepcopy(self._edge_label_kwargs)

        fig = fig if fig is not None else self.fig
        ax = ax if ax is not None else self.ax

        if positions is not None:
            self.parse_positions(**positions)

        self.parse_nodes(
            update_colormapping=update_colormapping,
            **(nodes if nodes else {}),
        )
        self.parse_edges(
            update_colormapping=update_colormapping,
            **(edges if edges else {}),
        )
        self.parse_node_labels(**(node_labels if node_labels else {}))
        self.parse_edge_labels(**(edge_labels if edge_labels else {}))
        self.mark_nodes(**(mark_nodes if mark_nodes else {}))
        self.mark_edges(**(mark_edges if mark_edges else {}))

        log.remark("Now drawing ...")

        # Draw nodes and edges
        self._mpl_nodes = nx.draw_networkx_nodes(
            self._g, pos=self.positions, ax=ax, **self._node_kwargs
        )
        self._mpl_edges = nx.draw_networkx_edges(
            self._g, pos=self.positions, ax=ax, **self._edge_kwargs
        )

        # NOTE networkx does not pass on the norms to the respective matplotlib
        #      functions. Hence, they need to be set manually. For the edges,
        #      the cmap also needs to be set manually. Can only be set for the
        #      edges if graph is undirected or `arrows=False`.
        self._mpl_nodes.set_norm(self._node_colormanager.norm)

        if not isinstance(self._mpl_edges, list):
            self._mpl_edges.set_norm(self._edge_colormanager.norm)
            self._mpl_edges.set_cmap(self._edge_colormanager.cmap)

        # Draw node labels and edge labels
        if self._show_node_labels:
            self._mpl_node_labels = nx.draw_networkx_labels(
                self._g,
                pos=self.positions,
                ax=ax,
                **self._node_label_kwargs,
            )
        if self._show_edge_labels:
            self._mpl_edge_labels = nx.draw_networkx_edge_labels(
                self._g,
                pos=self.positions,
                ax=ax,
                **self._edge_label_kwargs,
            )

        if not suppress_cbar:
            self.add_colorbars(
                show_node_cbar=self._show_node_cbar,
                show_edge_cbar=self._show_edge_cbar,
                fig=fig,
                ax=ax,
                **add_colorbars,
            )

        ax.axis("off")

        # Restore the previous drawing configuration
        self._node_colormanager = node_colormanager_cache
        self._edge_colormanager = edge_colormanager_cache
        self._node_cbar_kwargs = node_cbar_kwargs_cache
        self._edge_cbar_kwargs = edge_cbar_kwargs_cache
        self._show_node_cbar = show_node_cbar_cache
        self._show_edge_cbar = show_edge_cbar_cache
        self._show_node_labels = show_node_labels_cache
        self._show_edge_labels = show_edge_labels_cache
        self._node_kwargs = node_kwargs_cache
        self._edge_kwargs = edge_kwargs_cache
        self._node_label_kwargs = node_label_kwargs_cache
        self._edge_label_kwargs = edge_label_kwargs_cache

    def add_colorbars(
        self,
        *,
        show_node_cbar=True,
        show_edge_cbar=True,
        fig=None,
        ax=None,
        remove_previous=True,
        **update_cbar_kwargs,
    ):
        """Adds colorbars for the drawn nodes and edges.

        Args:
            show_node_cbar (bool, optional): Whether to create a colorbar for
                the nodes
            show_edge_cbar (bool, optional): Whether to create a colorbar for
                the edges
            fig (None, optional): matplotlib figure
            ax (None, optional): matplotlib axis
            remove_previous (bool, optional): Whether the colorbars which are
                currently associated with the ``GraphPlot`` are removed.
                If False, the GraphPlot still loses track of the colorbars,
                they can not be removed via the GraphPlot afterwards.
            **update_cbar_kwargs: Update both node and edge colorbar kwargs,
                passed to :py:meth:`~utopya.eval.plots._mpl.ColorManager.create_cbar`.
        """
        fig = fig if fig is not None else self.fig
        ax = ax if ax is not None else self.ax

        if show_node_cbar:
            if remove_previous and self._mpl_node_cbar:
                self._mpl_node_cbar.remove()

            self._mpl_node_cbar = self._node_colormanager.create_cbar(
                self._mpl_nodes,
                fig=fig,
                ax=ax,
                **self._node_cbar_kwargs,
                **update_cbar_kwargs,
            )

        if show_edge_cbar:
            if isinstance(self._mpl_edges, list):
                # When drawing arrows, draw_networkx_edges returns a list of
                # FancyArrowPatches which can not be used directly in
                # fig.colorbar, but needs conversion to a PatchCollection and
                # manual transfer of the chosen colormap and normalization ...
                edge_pc = mpl.collections.PatchCollection(self._mpl_edges)
                edge_pc.set_norm(self._edge_colormanager.norm)
                edge_pc.set_cmap(self._edge_colormanager.cmap)
            else:
                edge_pc = self._mpl_edges

            if remove_previous and self._mpl_edge_cbar:
                self._mpl_edge_cbar.remove()

            self._mpl_edge_cbar = self._edge_colormanager.create_cbar(
                edge_pc,
                fig=fig,
                ax=ax,
                **self._edge_cbar_kwargs,
                **update_cbar_kwargs,
            )

    def clear_plot(self, *, keep_colorbars: bool = False):
        """Removes all matplotlib objects associated with the GraphPlot from
        the respective axis. The GraphPlot loses track of all those objects,
        the respective class attributes are reset.

        Args:
            keep_colorbars (bool, optional): Whether to keep the node and edge
                colorbars. If True, the GraphPlot still loses track of the
                colorbars, they can not be removed via the GraphPlot
                afterwards.
        """
        # Remove colorbars. It is important to remove them before the
        # associated mappable.
        if not keep_colorbars:
            if self._mpl_node_cbar:
                try:
                    self._mpl_node_cbar.remove()
                except:
                    pass
            if self._mpl_edge_cbar:
                try:
                    self._mpl_edge_cbar.remove()
                except:
                    pass

        # Remove nodes and edges
        if self._mpl_nodes:
            self._mpl_nodes.remove()

        # Distinguish between FancyArrowPatch-list and LineCollection
        if isinstance(self._mpl_edges, list):
            for i in range(len(self._mpl_edges)):
                self._mpl_edges[i].remove()

        elif self._mpl_edges:
            self._mpl_edges.remove()

        # Remove labels
        if self._mpl_node_labels:
            for label in self._mpl_node_labels.values():
                label.remove()

        if self._mpl_edge_labels:
            for label in self._mpl_edge_labels.values():
                label.remove()

        # Reset class attributes
        self._mpl_node_cbar = None
        self._mpl_edge_cbar = None
        self._mpl_nodes = None
        self._mpl_edges = None
        self._mpl_node_labels = None
        self._mpl_edge_labels = None

        log.remark("Cleared GraphPlot.")

    # .........................................................................
    # Public helper methods

    @classmethod
    def from_group(
        cls,
        graph_group: "utopya.eval.groups.GraphGroup",
        *,
        graph_creation: dict = None,
        register_property_maps: dict = None,
        clear_existing_property_maps: bool = True,
        **init_kwargs,
    ):
        """Initializes a :py:class:`.GraphPlot` from a
        :py:class:`~utopya.eval.groups.GraphGroup`.

        Args:
            graph_group (utopya.eval.groups.GraphGroup): The graph group
            graph_creation (dict, optional): Configuration of the graph
                creation. Passed to :py:meth:`.create_graph_from_group`
            register_property_maps (dict, optional): Properties to be
                registered in the graph group before the graph creation keyed
                by name
            clear_existing_property_maps (bool, optional): Whether to clear any
                existing property maps from the graph group
            **init_kwargs: Passed to :py:meth:`.__init__`
        """
        g = cls.create_graph_from_group(
            graph_group,
            register_property_maps=register_property_maps,
            clear_existing_property_maps=clear_existing_property_maps,
            **(graph_creation if graph_creation is not None else {}),
        )

        return cls(g=g, **init_kwargs)

    @staticmethod
    def create_graph_from_group(
        graph_group: "utopya.eval.groups.GraphGroup",
        *,
        register_property_maps: dict = None,
        clear_existing_property_maps: bool = True,
        **graph_creation,
    ) -> nx.Graph:
        """Creates a :py:class:`networkx.Graph` from a
        :py:class`utopya.eval.groups.GraphGroup`.
        Additional property maps may be added to the group beforehand.

        Args:
            graph_group (utopya.eval.groups.GraphGroup): The group to create
                the graph from
            register_property_maps (dict, optional): Properties to be
                registered in the graph group before the graph creation keyed
                by name.
            clear_existing_property_maps (bool, optional): Whether to clear any
                existing property maps from the graph group.
            **graph_creation: Configuration of the graph creation.
                Passed on to the ``create_graph`` method implemented in dantro,
                :py:meth:`~dantro.groups.graph.GraphGroup.create_graph`.

        Returns:
            networkx.Graph: The created graph object.
        """
        # Register external property data
        if register_property_maps:
            # Clear existing property maps in order to not have side effects if
            # plotting multiple times, e.g. in interactive mode. This is
            # important because the graph_group most probably is a reference.
            if clear_existing_property_maps:
                graph_group.property_maps.clear()

            # Can register now
            for tag, pmap in register_property_maps.items():
                graph_group.register_property_map(tag, pmap)

        g = graph_group.create_graph(**graph_creation)
        return g

    def parse_positions(
        self,
        *,
        from_dict: Dict[Any, Tuple[float, float]] = None,
        model: Union[str, Callable] = None,
        **kwargs,
    ):
        """Parses the node positioning configuration. If a node positioning
        model is to be used, (re)calculates the positions.

        Args:
            from_dict (dict, optional): Explicit node positions (2-tuples keyed
                by node). If given, the ``model`` argument will be ignored.
            model (Union[str, Callable], optional): The model used for node
                positioning. If it is a string, it is looked up from the
                available networkx positioning models. If None, the spring
                model is used.
                If it is callable, it will be called with the graph as first
                positional argument.
            **kwargs: Passed to the node positioning routine

        Raises:
            ModuleNotFoundError: If a graphviz model was chosen but pygraphviz
                was not importable (via networkx).
        """
        # Set spring-layout as default if nothing else is specified
        if from_dict is None and model is None:
            model = "spring"

        if from_dict is not None:
            if model is not None:
                warnings.warn(
                    "Node positions were specified *both* via a positioning "
                    "model and explicitly via the `from_dict` argument. The "
                    "specified model will be ignored. To remove this warning, "
                    "set the graph_drawing.positions.model entry to None.",
                    UserWarning,
                )
            self.positions = copy.deepcopy(from_dict)

        else:
            log.remark(
                "Calculating the node positions using a positioning model ..."
            )

            if callable(model):
                self.positions = model(self._g, **kwargs)

            elif model.startswith("graphviz_"):
                try:
                    # graphviz models
                    model = model[9:]
                    self.positions = nx.nx_pydot.graphviz_layout(
                        self._g, prog=model, **kwargs
                    )

                except ModuleNotFoundError as err:
                    raise ModuleNotFoundError(
                        "When trying to use the graphviz node positioning "
                        f"model '{model}': '{err}'"
                    ) from err

            # else: is a networkx positioning model
            else:
                self.positions = POSITIONING_MODELS_NETWORKX[model](
                    self._g, **kwargs
                )

    def parse_nodes(
        self,
        *,
        node_size=None,
        node_color=None,
        alpha=None,
        cmap=None,
        cmap_norm=None,
        vmin: float = None,
        vmax: float = None,
        edgecolors=None,
        colorbar: dict = None,
        update_colormapping: bool = True,
        **kwargs,
    ):
        """Parses the node layout configuration and updates the node kwargs of
        the GraphPlot.

        The following arguments are available for property mapping:
        ``node_size``, ``node_color``, ``alpha``.

        Args:
            node_size (None, optional): Size of nodes (default=300). Available
                for property mapping. Can be mapped directly from the nodes'
                ``degree``, ``in_degree``, or ``out_degree`` by setting the
                ``from_property`` argument accordingly.
            node_color (None, optional): Single color (string or RGB(A) tuple
                or numeric value) or sequence of colors.
                If numeric values are specified they will be mapped to colors
                using the cmap and vmin, vmax parameters.

                If mapped from property it may contain an additional
                ``map_to_scalar``, which is a dict of numeric target values
                keyed by property value. This allows to map from non-numeric
                (e.g. categorical) properties.
            alpha (None, optional): The node transparency
            cmap (None, optional): The colormap. Passed as ``cmap`` to
                :py:class:`~utopya.eval.plots._mpl.ColorManager`.
            cmap_norm (None, optional): The norm used for the color mapping.
                Passed as ``norm`` to
                :py:class:`~utopya.eval.plots._mpl.ColorManager`.
                Is overwritten, if a discrete colormap is specified in
                ``cmap``.
            vmin (float, optional): Minimum for the colormap scaling
            vmax (float, optional): Maximum for the colormap scaling
            edgecolors (optional): Colors of node borders. The default is
                'none', i.e. no node border is drawn.
            colorbar (dict, optional): The node colorbar configuration.
                The following keys are available:

                enabled (bool, optional):
                    Whether to plot a colorbar. Enabled by default if
                    ``node_color`` is mapped from property.
                labels (dict, optional):
                    Colorbar tick-labels keyed by tick position (see
                    :py:class:`~utopya.eval.plots._mpl.ColorManager`).
                further kwargs: Passed on to
                    :py:meth:`~utopya.eval.plots._mpl.ColorManager.create_cbar`

            update_colormapping (bool, optional): Whether to reconfigure the
                nodes' :py:class:`~utopya.eval.plots._mpl.ColorManager`
                (default: True).
                If False, the respective arguments are ignored.
                Set to False if doing repetitive plotting with fixed
                colormapping.
            **kwargs: Update the node kwargs. Passed to
                :py:func:`~networkx.drawing.nx_pylab.draw_networkx_nodes`
                when calling :py:meth:`.draw`
        """
        if kwargs.pop("nodelist", None) is not None:
            warnings.warn(
                "The 'nodelist' argument will be ignored. To draw a subset of "
                "nodes use the 'select.nodelist' argument instead.",
                UserWarning,
            )

        # Update node kwargs with simple kwargs. All kwargs that might need
        # extra treatment are caught explicitly and handled below.
        self._node_kwargs.update(kwargs)

        self._node_kwargs["nodelist"] = self._nodes_to_draw

        # Do the property mapping ...
        # Node sizes
        if isinstance(node_size, dict) and "from_property" in node_size:
            prop = node_size["from_property"]
            interval = node_size.get("scale_to_interval", None)

            try:
                _node_sizes = np.array(
                    [self._g.nodes[n][prop] for n in self._nodes_to_draw]
                )
            except KeyError as err:
                if prop == "degree":
                    _node_sizes = node_sizes = np.array(
                        [self._g.degree[n] for n in self._nodes_to_draw]
                    )
                elif prop == "in_degree":
                    _node_sizes = node_sizes = np.array(
                        [self._g.in_degree[n] for n in self._nodes_to_draw]
                    )
                elif prop == "out_degree":
                    _node_sizes = node_sizes = np.array(
                        [self._g.out_degree[n] for n in self._nodes_to_draw]
                    )
                else:
                    raise ValueError(
                        f"No property {prop} found. Make sure the property "
                        "exists for all nodes to draw. Additional options: "
                        "degree, in_degree, out_degree"
                    ) from err

            # If there are nodes to be shrinked, set their size to zero
            to_shrink = np.isin(self._nodes_to_draw, self._nodes_to_shrink)
            _node_sizes[to_shrink] = 0
            _node_sizes[~to_shrink] = self._scale_to_interval(
                _node_sizes[~to_shrink], interval
            )
            self._node_kwargs["node_size"] = list(_node_sizes)

        elif node_size is not None:
            self._node_kwargs["node_size"] = node_size

        # Node colors
        if isinstance(node_color, dict) and "from_property" in node_color:
            prop = node_color["from_property"]
            interval = node_color.get("scale_to_interval", None)

            # If provided a mapping, map the property values to scalar values
            if "map_to_scalar" in node_color:
                map_to_scalar = np.vectorize(node_color["map_to_scalar"].get)
                _node_colors = list(map_to_scalar(_node_colors))

            else:
                _node_colors = self._scale_to_interval(
                    [self._g.nodes[n][prop] for n in self._nodes_to_draw],
                    interval,
                )

            self._node_kwargs["node_color"] = _node_colors
            self._show_node_cbar = True

        elif node_color is not None:
            self._node_kwargs["node_color"] = node_color

        # Node transparency
        if isinstance(alpha, dict) and "from_property" in alpha:
            prop = alpha["from_property"]
            interval = alpha.get("scale_to_interval", None)

            self._node_kwargs["alpha"] = self._scale_to_interval(
                [self._g.nodes[n][prop] for n in self._nodes_to_draw], interval
            )

        elif alpha is not None:
            self._node_kwargs["alpha"] = alpha

        # ... property mapping done.

        # If a single or no value was set for the node size, create an explicit
        # node size mapping if there are nodes to be shrinked to size zero.
        # If no size was given, use the default size 300.
        if self._nodes_to_shrink and not isinstance(
            self._node_kwargs.get("node_size", None), list
        ):
            _size = self._node_kwargs.get("node_size", 300)
            self._node_kwargs["node_size"] = [
                0 if n in self._nodes_to_shrink else _size
                for n in self._nodes_to_draw
            ]

        # Update the edgecolors of the nodes. Set the edgecolor to 'none'
        # (=transparent) if it has not been set yet.
        if edgecolors is not None:
            self._node_kwargs["edgecolors"] = edgecolors

        elif "edgecolors" not in self._node_kwargs:
            self._node_kwargs["edgecolors"] = "none"

        # For directed graphs (arrows enabled) individual FancyArrowPatches are
        # drawn. Make sure that the arrows end right at the node borders.
        if self._g.is_directed():
            self._configure_node_patch_sizes()

        # Get colorbar configuration and update the existing kwargs
        colorbar = colorbar if colorbar is not None else {}
        cbar_labels = colorbar.pop("labels", None)
        self._show_node_cbar = colorbar.pop("enabled", self._show_node_cbar)
        self._node_cbar_kwargs.update(colorbar)

        # (Re-)create the colormanager. Only do so if there is no colormanager
        # yet or if any kwargs are given.
        if update_colormapping and (
            self._node_colormanager is None
            or any(
                [
                    arg is not None
                    for arg in (cmap, cmap_norm, vmin, vmax, cbar_labels)
                ]
            )
        ):
            # Replace all kwargs that are None with the current configuration.
            # That way the current configuration is not lost but updated.
            if vmin is None:
                vmin = self._node_kwargs.get("vmin", None)
            if vmax is None:
                vmax = self._node_kwargs.get("vmax", None)
            if self._node_colormanager is not None:
                if cmap is None:
                    cmap = self._node_colormanager.cmap
                if cmap_norm is None:
                    cmap_norm = self._node_colormanager.norm
                if cbar_labels is None:
                    cbar_labels = self._node_colormanager.labels

            # Set up the ColorManager
            self._node_colormanager = ColorManager(
                cmap=cmap,
                norm=cmap_norm,
                labels=cbar_labels,
                vmin=vmin,
                vmax=vmax,
            )
            self._node_kwargs["cmap"] = self._node_colormanager.cmap

    def parse_edges(
        self,
        *,
        width=None,
        edge_color=None,
        edge_cmap=None,
        cmap_norm=None,
        edge_vmin: float = None,
        edge_vmax: float = None,
        colorbar: dict = None,
        update_colormapping: bool = True,
        **kwargs,
    ):
        """Parses the edge layout configuration and updates the edge kwargs of
        the GraphPlot.

        The following arguments are available for property mapping:
        ``width``, ``edge_color``.

        Args:
            width (None, optional): Line width of edges
            edge_color (None, optional): Single color (string or RGB(A) tuple
                or numeric value) or sequence of colors (default='k').
                If numeric values are specified they will be mapped to colors
                using the edge_cmap and edge_vmin, edge_vmax parameters.

                If mapped from property it may contain an additional
                ``map_to_scalar``, which is a dict of numeric target values
                keyed by property value. This allows to map from non-numeric
                (e.g. categorical) properties.
            edge_cmap (None, optional): The colormap. Passed as ``cmap`` to
                :py:class:`~utopya.eval.plots._mpl.ColorManager`.
            cmap_norm (None, optional): The norm used for the color mapping.
                Passed as ``norm`` to
                :py:class:`~utopya.eval.plots._mpl.ColorManager`.
                Is overwritten, if a discrete colormap is specified in
                ``edge_cmap``.
                If arrows are to be drawn (i.e. for directed edges with
                arrows=True), only norms of type matplotlib.colors.Normalize
                are allowed.
            edge_vmin (float, optional): Minimum for the colormap scaling
            edge_vmax (float, optional): Maximum for the colormap scaling
            colorbar (dict, optional): The edge colorbar configuration.
                The following keys are available:

                enabled (bool, optional):
                    Whether to plot a colorbar. Enabled by default if
                    ``edge_color`` is mapped from property.
                labels (dict, optional):
                    Colorbar tick-labels keyed by tick position (see
                    :py:class:`~utopya.eval.plots._mpl.ColorManager`).
                further kwargs: Passed on to
                    :py:meth:`~utopya.eval.plots._mpl.ColorManager.create_cbar`

            update_colormapping (bool, optional): Whether to reconfigure the
                edges' :py:class:`~utopya.eval.plots._mpl.ColorManager`
                (default=True).
                If False, the respective arguments are ignored.
                Set to False if doing repetitive plotting with fixed
                colormapping.
            **kwargs: Update the edge kwargs. Passed to
                :py:func:`~networkx.drawing.nx_pylab.draw_networkx_nodes`
                when calling :py:meth:`.draw`.

        Raises:
            TypeError: On norm type other than
                :py:class:`matplotlib.colors.Normalize` and if arrows are to be
                drawn.
        """
        if kwargs.pop("edgelist", None) is not None:
            warnings.warn(
                "The 'edgelist' argument will be ignored. Use the select "
                "configuration to specify a subgraph to be drawn.",
                UserWarning,
            )
        self._edge_kwargs["edgelist"] = self._edges_to_draw

        # Update edge kwargs with simple kwargs. All kwargs that might need
        # extra treatment are caught explicitly and handled below.
        self._edge_kwargs.update(kwargs)

        # Do the property mapping ...
        # Edge width
        if isinstance(width, dict) and "from_property" in width:
            prop = width["from_property"]
            interval = width.get("scale_to_interval", None)

            self._edge_kwargs["width"] = self._scale_to_interval(
                [self._g.edges[n][prop] for n in self._edges_to_draw], interval
            )

        elif width is not None:
            self._edge_kwargs["width"] = width

        # Node colors
        if isinstance(edge_color, dict) and "from_property" in edge_color:
            prop = edge_color["from_property"]
            interval = edge_color.get("scale_to_interval", None)

            # If provided a mapping, map the property values to scalar values
            if "map_to_scalar" in edge_color:
                map_to_scalar = np.vectorize(edge_color["map_to_scalar"].get)
                _edge_colors = list(map_to_scalar(_edge_colors))

            else:
                _edge_colors = self._scale_to_interval(
                    [self._g.edges[n][prop] for n in self._edges_to_draw],
                    interval,
                )

            self._edge_kwargs["edge_color"] = _edge_colors
            self._show_edge_cbar = True

        elif edge_color is not None:
            self._edge_kwargs["edge_color"] = edge_color

        # ... property mapping done.

        # Get colorbar configuration and update the existing kwargs
        colorbar = colorbar if colorbar is not None else {}
        cbar_labels = colorbar.pop("labels", None)
        self._show_edge_cbar = colorbar.pop("enabled", self._show_edge_cbar)
        self._edge_cbar_kwargs.update(colorbar)

        # (Re-)create the colormanager. Only do so if there is no colormanager
        # yet or if any kwargs are given.
        if update_colormapping and (
            self._edge_colormanager is None
            or any(
                [
                    arg is not None
                    for arg in (
                        edge_cmap,
                        cmap_norm,
                        edge_vmin,
                        edge_vmax,
                        cbar_labels,
                    )
                ]
            )
        ):
            # Replace all kwargs that are None with the current configuration.
            # That way the current configuration is not lost but updated.
            if edge_vmin is None:
                edge_vmin = self._edge_kwargs.get("edge_vmin", None)
            if edge_vmax is None:
                edge_vmax = self._edge_kwargs.get("edge_vmax", None)
            if self._edge_colormanager is not None:
                if edge_cmap is None:
                    edge_cmap = self._edge_colormanager.cmap
                if cmap_norm is None:
                    cmap_norm = self._edge_colormanager.norm
                if cbar_labels is None:
                    cbar_labels = self._edge_colormanager.labels

            # Set up the ColorManager
            self._edge_colormanager = ColorManager(
                cmap=edge_cmap,
                norm=cmap_norm,
                labels=cbar_labels,
                vmin=edge_vmin,
                vmax=edge_vmax,
            )
            self._edge_kwargs["edge_cmap"] = self._edge_colormanager.cmap

        # NOTE In draw_networkx_edges, the Normalize norm is applied
        #      explicitly. Since the norm cannot be updated later (as edges
        #      with arrows are FancyArrowPatches), other norms than Normalize
        #      are forbidden here. When mapping the colors using the
        #      Colormanager a colorbar drawn from the edges mappable would
        #      still be wrong.
        if (
            isinstance(self._g, nx.DiGraph)
            and self._edge_kwargs.get("arrows", True)
            and type(self._edge_colormanager.norm) != mpl.colors.Normalize
        ):
            raise TypeError(
                "Received invalid norm type: "
                f"{type(self._edge_colormanager.norm)}. For directed edges"
                " with 'arrows=True', only the matplotlib.colors.Normalize"
                " base class is supported."
            )

    def parse_node_labels(
        self,
        *,
        enabled: bool = False,
        show_only: list = None,
        labels: dict = None,
        format: str = "{label}",
        decode: str = None,
        **kwargs,
    ):
        """Parses the node labels configuration and updates the node label
        kwargs of the :py:class:`.GraphPlot`.

        Args:
            enabled (bool, optional): Whether to draw node labels.
            show_only (list, optional): If given, labels are drawn only for the
                nodes in this list
            labels (dict, optional): Custom text labels keyed by node.
                Available for property mapping.
            format (str, optional): If ``labels`` are mapped from property this
                format string containing a ``label`` key is used for all node
                labels.
            decode (str, optional): Decoding specifier which is applied to all
                property values if ``format`` is used.
            **kwargs: Update the node label kwargs. Passed to
                nx.draw_networkx_labels when calling
                :py:meth:`.draw`.
        """
        if enabled:
            self._show_node_labels = True

        # Update node label kwargs with simple kwargs. All kwargs that might
        # need extra treatment are caught explicitly and handled below.
        self._node_label_kwargs.update(kwargs)

        if not enabled and not labels:
            # Don't update the node_labels.labels -> can stop here
            return

        def to_show(n):
            if show_only is None:
                return n not in self._nodes_to_shrink
            else:
                return n in show_only and n not in self._nodes_to_shrink

        if labels:
            if "from_property" in labels:
                prop = labels["from_property"]
                _labels = {
                    n: format.format(
                        label=(
                            self._g.nodes[n][prop]
                            if decode is None
                            else self._g.nodes[n][prop].decode(decode)
                        )
                    )
                    for n in self._nodes_to_draw
                    if to_show(n)
                }

            else:
                _labels = copy.deepcopy(labels)

                if show_only is not None:
                    # show_only takes precedence over the provided node labels.
                    # Keep only those that are in show_only. nodes in show_only
                    # for which no label is given are labeled with their index.
                    for n in labels.keys():
                        if not to_show(n):
                            del _labels[n]

                    for n in show_only:
                        if (
                            n not in _labels.keys()
                            and n not in self._nodes_to_shrink
                        ):
                            _labels[n] = n

        elif enabled:
            # If enabled but no labels given, label nodes with their index.
            _labels = {n: n for n in self._nodes_to_draw if to_show(n)}

        self._node_label_kwargs["labels"] = _labels

    def parse_edge_labels(
        self,
        *,
        enabled: bool = False,
        show_only: list = None,
        edge_labels: dict = None,
        format: str = "{label}",
        decode: str = None,
        **kwargs,
    ):
        """Parses the edge labels configuration and updates the edge label
        kwargs of the GraphPlot.

        Args:
            enabled (bool, optional): Whether to draw edge labels.
            show_only (list, optional): If given, labels are drawn only for the
                edges (2-tuples) in this list
            edge_labels (dict, optional): Custom text labels keyed by edge
                (2-tuple). Available for property mapping.
            format (str, optional): If ``edge_labels`` are mapped from property
                this format string containing a ``label`` key is used for all
                edge labels.
            decode (str, optional): Decoding specifier which is applied to all
                property values if ``format`` is used.
            **kwargs: Update the edge label kwargs. Passed to
                nx.draw_networkx_edge_labels when calling
                :py:meth:`.draw`.
        """
        # Catch a pitfall: There is no 'labels' argument for the edge labels
        # (as there is for the node labels), here it is named 'edge_labels'.
        if "labels" in kwargs:
            raise ValueError(
                "Received invalid 'labels' key in edge label configuration. "
                "For specifying an edge label dict, use the key 'edge_labels'."
            )

        if enabled:
            self._show_edge_labels = True

        # Update edge label kwargs with simple kwargs. All kwargs that might
        # need extra treatment are caught explicitly and handled below.
        self._edge_label_kwargs.update(kwargs)

        if not enabled and not edge_labels:
            # Don't update the edge_labels.labels -> can stop here
            return

        def to_show(e):
            return True if show_only is None else e[:2] in show_only

        if show_only is not None:
            show_only = [tuple(e) for e in show_only]

        if edge_labels:
            if "from_property" in edge_labels:
                prop = edge_labels["from_property"]
                _labels = {
                    e[:2]: format.format(
                        label=(
                            self._g.edges[e][prop]
                            if decode is None
                            else self._g.edges[e][prop].decode(decode)
                        )
                    )
                    for e in self._edges_to_draw
                    if to_show(e)
                }

            else:
                _labels = copy.deepcopy(edge_labels)

                if show_only is not None:
                    # show_only takes precedence over the provided edge labels.
                    # Keep only those that are in show_only. edges in show_only
                    # for which no label is given are labeled with their
                    # (source, destination) pair.
                    for e in edge_labels.keys():
                        if not to_show(e):
                            del _labels[e]

                    for e in show_only:
                        if e not in _labels.keys():
                            _labels[e] = e

        elif enabled:
            # If enabled but no labels given, label edges with their
            # (source, destination) pair.
            _labels = {e[:2]: e[:2] for e in self._edges_to_draw if to_show(e)}

        self._edge_label_kwargs["edge_labels"] = _labels

    def mark_nodes(
        self, *, nodelist: list = None, color=None, colors: dict = None
    ):
        """Mark specific nodes by changing their edgecolor.

        .. note::

            This function overwrites the ``edgecolors`` entry in the node
            kwargs. Thus it might overwrite an existing ``edgecolors`` entry
            specified via
            :py:meth:`.parse_nodes`
            (and vice versa).

        Args:
            nodelist (list, optional): Nodes to mark with the color specified
                via ``color``
            color (None, optional): Single edgecolor to use for the nodes in
                ``nodelist``
            colors (dict, optional): Edgecolors keyed by node to mark. Must be
                None if ``nodelist`` is given.

        Raises:
            ValueError: On ambiguous or missing mark_nodes configuration
        """
        if nodelist is None and color is None and colors is None:
            return

        if colors is not None:
            if not (nodelist is None and color is None):
                raise ValueError(
                    "Received invalid 'mark_nodes' kwargs. Provide _either_ a "
                    "'colors' dict only _or_ an 'nodelist' together with a "
                    "single 'color'."
                )

        elif (nodelist is None) != (color is None):
            _missing_arg = "nodelist" if nodelist is None else "color"
            raise ValueError(
                f"Missing argument '{_missing_arg}' in 'mark_nodes' kwargs."
            )

        if self._node_kwargs.get("edgecolors", None) is None:
            # There is no edgecolors entry yet. Use the current node colors
            # as base colors. If there are none, use the default networkx node
            # color.
            base_color = self._node_kwargs.get("node_color", "none")
        else:
            # Use the existing edgecolors entry.
            base_color = self._node_kwargs["edgecolors"]

        # From the base color(s) create dict of colors keyed by node
        if mpl.colors.is_color_like(base_color):
            _colors = {n: base_color for n in self._nodes_to_draw}
        else:
            # ... must be a list of colors.
            # If base_color contains numeric values, they need to be
            # transformed via the specified colormap.
            if not mpl.colors.is_color_like(base_color[0]):
                base_color = self._node_colormanager.map_to_color(base_color)

            _colors = {
                n: base_color[i] for i, n in enumerate(self._nodes_to_draw)
            }

        # Update the color dict with the values from the mark configuration
        if nodelist:
            for n in nodelist:
                _colors[n] = color
        else:
            _colors.update(colors)

        # The color values are aligned with 'self._nodes_to_draw' since
        # dictionaries don't change their ordering.
        self._node_kwargs["edgecolors"] = list(_colors.values())

        # Reconfigure the node patch sizes which increase if a node boundary
        # is drawn.
        if self._g.is_directed():
            self._configure_node_patch_sizes()

    def mark_edges(
        self, *, edgelist: list = None, color=None, colors: dict = None
    ):
        """Mark specific edges by changing their color.

        .. note::

            This function overwrites the ``edge_color`` entry in the edge
            kwargs. Thus it might overwrite an existing ``edge_color`` entry
            specified via
            :py:meth:`.parse_edges`
            (and vice versa).

        Args:
            edgelist (list, optional): Edges to mark with the color specified
                via ``color``
            color (None, optional): Single color to use for the edges in
                ``edgelist``
            colors (dict, optional): Colors keyed by edge (2-tuple) to mark.
                Must be None if ``edgelist`` is given.

        Raises:
            ValueError: On ambiguous or missing mark_edges configuration
        """
        if edgelist is None and color is None and colors is None:
            return

        if colors is not None:
            if not (edgelist is None and color is None):
                raise ValueError(
                    "Received invalid 'mark_edges' kwargs. Provide _either_ a "
                    "'colors' dict only _or_ an 'edgelist' together with a "
                    "single 'color'."
                )

        elif (edgelist is None) != (color is None):
            _missing_arg = "edgelist" if edgelist is None else "color"
            raise ValueError(
                f"Missing argument '{_missing_arg}' in 'mark_edges' kwargs."
            )

        # Create dict of colors keyed by node based on the current edge_color.
        # If there is none, use the default networkx edge color.
        base_color = self._edge_kwargs.get("edge_color", "k")

        if mpl.colors.is_color_like(base_color):
            _colors = {e[:2]: base_color for e in self._edges_to_draw}
        else:
            # ... must be a list of colors. Transform to color-like if needed.
            if not mpl.colors.is_color_like(base_color[0]):
                base_color = self._edge_colormanager.map_to_color(base_color)

            _colors = {
                e[:2]: base_color[i] for i, e in enumerate(self._edges_to_draw)
            }

        # Update the color dict with the values from the mark configuration
        if edgelist:
            for e in edgelist:
                e = tuple(e)
                if not isinstance(self._g, nx.DiGraph) and e not in _colors:
                    e = e[::-1]
                _colors[e] = color
        else:
            for e, c in colors.items():
                if not isinstance(self._g, nx.DiGraph) and e not in _colors:
                    e = e[::-1]
                _colors[e] = c

        self._edge_kwargs["edge_color"] = list(_colors.values())

    # .........................................................................
    # Private helper methods

    def _configure_node_patch_sizes(self):
        """In the edge drawing kwargs adjusts the node patch specifications
        ensuring that, in the case of a directed graph, the arrows end exactly
        at the node boundaries. Besides the ``node_size`` also accounts for
        node boundaries.
        """
        # NOTE This also implies that edges can only be drawn if both their
        #      source and destination are drawn (as long as everything is based
        #      on a single _nodes_to_draw list).
        self._edge_kwargs["nodelist"] = self._nodes_to_draw

        if "node_shape" in self._node_kwargs:
            self._edge_kwargs["node_shape"] = self._node_kwargs["node_shape"]

        patch_size = self._node_kwargs.get("node_size", 300)
        edgecolors = self._node_kwargs.get("edgecolors", "none")

        # If node boundaries are drawn the size of the node patch increases
        # depending on the boundary width. Adjust the patch size accordingly.
        # Only do the patch size configuration if 'edgecolors' is not 'none'
        # (=no patch boundary).
        if edgecolors != "none":
            lw = self._node_kwargs.get(
                "linewidths", mpl.rcParams["lines.linewidth"]
            )

            if not isinstance(patch_size, (int, float)):
                patch_size = np.array(patch_size)

            if not isinstance(lw, (int, float)):
                lw = np.array(lw)

            # Adjust the patch size wherever a node boundary is to be drawn
            if mpl.colors.is_color_like(edgecolors):
                patch_size = (np.sqrt(patch_size) + lw) ** 2

            else:
                patch_size = np.where(
                    [ec == "none" for ec in edgecolors],
                    patch_size,
                    (np.sqrt(patch_size) + lw) ** 2,
                )

        self._edge_kwargs["node_size"] = patch_size

    @staticmethod
    def _scale_to_interval(data: list, interval=None) -> list:
        """Rescales the data linearly to the given interval. If no interval is
        given the data is returned as it is.

        Args:
            data (list): data that is rescaled linearly to the given interval
            interval (optional): The target interval

        Returns:
            list: rescaled data

        Raises:
            ValueError: On invalid interval specification
        """
        if interval is None:
            return data

        try:
            lim_low, lim_up = interval
        except ValueError as err:
            raise ValueError(
                "'interval' must be a 2-tuple or list of length 2! Received: "
                f"{interval}"
            ) from err

        data = np.array(data)
        max_val = np.max(data)
        min_val = np.min(data)

        if max_val > min_val:
            rescaled_data = (data - min_val) / (max_val - min_val) * (
                lim_up - lim_low
            ) + lim_low
        else:
            # If all values are equal, set them to the mean of the interval
            rescaled_data = np.zeros_like(data) + (lim_up - lim_low) / 2.0

        return list(rescaled_data)

    def _select_subgraph(
        self, nodelist: list = None, drop: bool = True, **kwargs
    ):
        """Select a subgraph to draw. Sets the lists of nodes and edges to draw
        and the nodes to shrink. Either a list of nodes is selected or radial
        selection is done.

        Args:
            nodelist (list, optional): If given, select nodes from list
            drop (bool, optional): Whether to remove the non-selected nodes and
                edges from the graph
            **kwargs: Passed to the selection routine
        """

        def select_from_list(*, nodelist: list, open_edges: bool = False):
            """Given a list of nodes, selects all nodes and edges needed for
            the graph drawing. If ``open_edges=False``, only those edges are
            selected for which both ends are in ``nodes``.

            Args:
                nodelist (list): Nodes to be selected
                open_edges (bool, optional): Whether to draw loose edges

            Returns:
                Tuple containing list of selected nodes, list of selected
                edges, and list of nodes to be shrinked to size zero.
            """
            subgraph = nx.induced_subgraph(self._g, nodelist)

            if open_edges:
                # Create outer subgraph from given nodes and all neighbors
                node_selection = set(nodelist)
                outer_nodes = set()

                for n in nodelist:
                    outer_nodes.update(nx.all_neighbors(self._g, n))

                outer_nodes -= node_selection
                node_selection = node_selection.union(outer_nodes)
                subgraph_outer = nx.induced_subgraph(self._g, node_selection)

                # Set of nodes to shrink is the difference of the two node sets
                nodes_to_shrink = list(subgraph_outer.nodes - set(nodelist))

                # From the outer subgraph remove edges between outer nodes
                edges_to_plot = (
                    subgraph_outer.edges
                    - nx.induced_subgraph(self._g, outer_nodes).edges
                )

                return (
                    list(subgraph_outer.nodes),
                    list(edges_to_plot),
                    nodes_to_shrink,
                )

            return list(subgraph.nodes), list(subgraph.edges), []

        def select_radial(
            *, center: int, radius: int, open_edges: bool = False
        ):
            """Selects all nodes around a given center within a given radius
            (measured in numbers of neighborhoods). If ``open_edges=False``, those
            edges are selected for which both ends are in the set of selected
            nodes.

            Args:
                center (int): Index of the central node
                radius (int): Selection radius
                open_edges (bool, optional): Whether to draw loose edges

            Returns:
                Tuple containing list of selected nodes, list of selected
                edges, and list of nodes to be shrinked to size zero.
            """
            # After num_nodes-1 iterations (below), all nodes would be selected
            if radius > self._g.number_of_nodes() - 1:
                radius = self._g.number_of_nodes() - 1

            # Identify nodes within the given radius around the central node.
            # Start by adding the central nodes and all its neighbors to a set.
            # Then, iteratively add all neighbors of the previously added nodes
            # to the set, until the given radius is reached.
            # TODO It might be worth testing the computational efficiency of this
            #      (also for large subgraphs) as each node is tried to be added to
            #      the set at least two times.
            # Store the current node selection
            node_selection = {center}
            # Store the nodes added to the selection in the previous step
            nbs_prev = {center}
            # Store the new nodes to be selected
            nbs_new = set()

            for i in range(radius):
                for n in nbs_prev:
                    nbs_new.update(nx.all_neighbors(self._g, n))

                nbs_prev = nbs_new - node_selection
                node_selection = node_selection.union(nbs_new)
                nbs_new.clear()

            if open_edges:
                # Create inner subgraph from all nodes within the given radius
                subgraph_inner = nx.induced_subgraph(self._g, node_selection)

                # Create outer subgraph from all nodes within radius=radius+1
                for n in nbs_prev:
                    nbs_new.update(nx.all_neighbors(self._g, n))

                outer_nodes = nbs_new - node_selection
                node_selection = node_selection.union(nbs_new)
                subgraph_outer = nx.induced_subgraph(self._g, node_selection)

                # Set of nodes to shrink is the difference of the two node sets
                nodes_to_shrink = list(
                    subgraph_outer.nodes - subgraph_inner.nodes
                )

                # From the outer subgraph remove edges between outer nodes
                edges_to_plot = (
                    subgraph_outer.edges
                    - nx.induced_subgraph(self._g, outer_nodes).edges
                )

                # Return the inner subgraph nodes and the outer subgraph edges
                return (
                    list(subgraph_outer.nodes),
                    list(edges_to_plot),
                    nodes_to_shrink,
                )

            subgraph = nx.induced_subgraph(self._g, node_selection)
            return list(subgraph.nodes), list(subgraph.edges), []

        # Select the nodes and edges to be drawn. The selection is always based
        # on the original graph self._g.
        if nodelist is None and not kwargs:
            # If no selection was specified, select all nodes and edges
            nodes_to_draw, edges_to_draw, nodes_to_shrink = (
                list(self._g.nodes),
                list(self._g.edges),
                [],
            )
        elif nodelist is not None:
            # Select from list of nodes
            nodes_to_draw, edges_to_draw, nodes_to_shrink = select_from_list(
                nodelist=nodelist, **kwargs
            )
        else:
            # Perform radial selection
            nodes_to_draw, edges_to_draw, nodes_to_shrink = select_radial(
                **kwargs
            )

        if drop:
            # Remove the nodes that are not selected. This also removes all
            # edges for which the source or destination was removed.
            nodes_to_remove = self._g.nodes - set(nodes_to_draw)
            if nodes_to_remove:
                self._g.remove_nodes_from(nodes_to_remove)

        self._nodes_to_draw = nodes_to_draw
        self._edges_to_draw = edges_to_draw
        self._nodes_to_shrink = nodes_to_shrink



"""Utility functions that assist plot functions"""

import itertools
import logging
from typing import Dict, Tuple

import numpy as np
import xarray as xr

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


def calc_pxmap_rectangles(
    *,
    x_coords: np.ndarray,
    y_coords: np.ndarray,
    x_scale: str = "lin",
    y_scale: str = "lin",
    default_pos: Tuple[float, float] = (0.0, 0.0),
    default_distance: float = 1.0,
    size_factor: float = 1.0,
    extend: bool = False,
) -> Tuple[xr.Dataset, Dict[str, tuple]]:
    """Calculates the positions and sizes of rectangles centered at the given
    coordinates in such a way that they fully cover the 2D space spanned by
    the coordinates.

    The exact values of the coordinates are arbitrary, but they should be
    ordered. The resulting rectangles will not necessarily be centered around
    the coordinate, but the distance between rectangle edges is set such that
    it lies halfway to the next coordinate in that dimension; the "halfway" is
    evaluated according to a certain scale.

    Args:
        x_coords (numpy.ndarray): The x coordinates
        y_coords (numpy.ndarray): The y coordinates
        x_scale (str, optional): The x-axis scale, used to determine rectangle
            sizes.
        y_scale (str, optional): The y-axis scale, used to determine rectangle
            sizes.
        default_pos (Tuple[float, float], optional): If any of the coordinates
            is not given, this information will be used for creating the
            rectangle specification.
        default_distance (float, optional): If any of the coordinates is not
            given or of size 1, this distance to the next data point is assumed
            in that dimension. In such a case, the scale does not have an
            effect on the resulting rectangle.
        size_factor (float, optional): Scaling factor for the rectangle sizes
        extend (bool, optional): Whether to extend the rectangles of the points
            at the border of the domain such that the coordinate is _not_ at
            the edge of the rectangle but in the bulk.

    Returns:
        Tuple[xarray.Dataset, Dict[str, tuple]]: The first tuple element is the
            dataset of rectangle specifications, each available as a data
            variable:

                - ``pos_x``, ``pos_y``: position of the lower-value coordinate
                  of the rectangle. Together, this specifies the bottom left-
                  hand corner of the rectangle (in a right-hand coordinate
                  system)
                - ``len_x``, ``len_y``: lengths of the rectangle sides in the
                  specified dimensions. Adding both these to the position leads
                  to the top-right hand corner of the rectangle
                - ``rect_spec``: A matplotlib-compatible rectangle
                  specification, i.e. (position, width, height)

            The second tuple element are the limits in x and y direction, given
            as dict with keys x and y.
    """

    # Define some allowed linear and logarithmic scales
    LIN_SCALES = ["lin", "linear"]
    LOG_SCALES = ["log", "symlog"]
    CATEGORIAL_SCALES = ["cat", "categorical", "cat_at_value"]

    # Helper functions ........................................................

    def determine_diffs_and_scale(
        coords: np.ndarray, *, scale: str, axis_name: str
    ) -> Tuple[np.ndarray, str]:
        """Given the coordinates and the scale, determine the distances
        between coordinates in one dimension. This is also used to check the
        scale.

        Args:
            coords (array-like): Sequence of coordinates
            scale (str): The scale to use
            axis_name (str): The name of the axis; used for error message only

        Returns:
            Tuple[numpy.ndarray, str]: Coordinate distances and the new scale

        Raises:
            ValueError: On invalid scale argument
        """
        if scale in LIN_SCALES:
            diffs = np.diff(coords)

        elif scale in LOG_SCALES:
            diffs = np.diff(np.log(coords))

        elif scale in CATEGORIAL_SCALES:
            # For categorial scale at value, use just some arbitrary integeres
            # for the coordinates; the scale can then be assumed linear.
            if scale != "cat_at_value":
                coords = list(range(len(coords)))
            diffs = np.diff(coords)
            scale = "lin"

        else:
            raise ValueError(
                "Invalid scale argument for {} axis: '{}'! "
                "Expected one of: {}"
                "".format(
                    scale,
                    axis_name,
                    ", ".join(LIN_SCALES + LOG_SCALES + CATEGORIAL_SCALES),
                )
            )

        return diffs, scale

    def expand_diffs(diffs, *, extend: bool, default: float) -> np.ndarray:
        """Adds the boundary values to the given diffs"""
        if diffs.size < 1:
            # Need to use the default value
            diffs = np.array([default, default])

        elif extend:
            # Add the first item to the beginning and the last to the end of
            # the diffs sequences to account for the borders
            diffs = np.insert(diffs, 0, diffs[0])
            diffs = np.append(diffs, diffs[-1])

        else:
            # No extension of borders desired --> add zeros to diffs
            diffs = np.insert(diffs, 0, 0.0)
            diffs = np.append(diffs, 0.0)

        return diffs

    def determine_limits(*, coords, sizes, scale: str) -> Tuple[float, float]:
        """Determine the limits of the space that is covered by the rectangles
        at the given coordinates and sizes.
        """
        if scale in LIN_SCALES:
            return (float(coords[0]) - sizes[0], float(coords[-1]) + sizes[-1])

        elif scale in LOG_SCALES:
            return (
                np.exp(np.log(float(coords[0])) - sizes[0]),
                np.exp(np.log(float(coords[-1])) + sizes[-1]),
            )

        # NOTE This point is not reached; scale is checked elsewhere

    # The functions below create a rectangle specifier given the x and y
    # positions and the desired sizes.
    # Have 4 explicit functions (one for each scale specification) because it's
    # very tedious to create the rectangle specifier in a general way ... Grml.

    def calc_linlin_rect(
        x: float, y: float, x_sizes: tuple, y_sizes: tuple
    ) -> tuple:
        """
        Create the rectangle specification for a single rectangle embedded in
        a lin-lin plot.

        Args:
            x (float): The x-coordinate
            y (float): The y-coordinate
            x_sizes (tuple): The distance to the rectangle edges in x direction
            y_sizes (tuple): The distance to the rectangle edges in y direction

        Returns:
            tuple: (x pos. bottom left-hand corner,
                    y pos. bottom left-hand corner,
                    total x length,
                    total y length)
        """
        return (
            x - x_sizes[0],
            y - y_sizes[0],
            x_sizes[0] + x_sizes[1],
            y_sizes[0] + y_sizes[1],
        )

    def calc_linlog_rect(x, y, x_sizes, y_sizes) -> tuple:
        """See calc_linlin_rect"""
        return (
            x - x_sizes[0],
            np.exp(np.log(y) - y_sizes[0]),
            x_sizes[0] + x_sizes[1],
            np.exp(np.log(y) + y_sizes[1]) - np.exp(np.log(y) - y_sizes[0]),
        )

    def calc_loglin_rect(x, y, x_sizes, y_sizes) -> tuple:
        """See calc_linlin_rect"""
        return (
            np.exp(np.log(x) - x_sizes[0]),
            y - y_sizes[0],
            np.exp(np.log(x) + x_sizes[1]) - np.exp(np.log(x) - x_sizes[0]),
            y_sizes[0] + y_sizes[1],
        )

    def calc_loglog_rect(x, y, x_sizes, y_sizes) -> tuple:
        """See calc_linlin_rect"""
        return (
            np.exp(np.log(x) - x_sizes[0]),
            np.exp(np.log(y) - y_sizes[0]),
            np.exp(np.log(x) + x_sizes[1]) - np.exp(np.log(x) - x_sizes[0]),
            np.exp(np.log(y) + y_sizes[1]) - np.exp(np.log(y) - y_sizes[0]),
        )

    # Create a map to select the callable via `in LIN_SCALES` boolean key pair
    FUNC_MAP = {
        # x lin?  y lin?
        (True, True): calc_linlin_rect,
        (True, False): calc_linlog_rect,
        (False, True): calc_loglin_rect,
        (False, False): calc_loglog_rect,
    }

    # .........................................................................

    log.debug("Calculating rectangle positions and sizes ...")

    # Allow None, in which case the rectangle is set to a default position on
    # that axis
    x_coords = x_coords if x_coords is not None else [default_pos[0]]
    y_coords = y_coords if y_coords is not None else [default_pos[1]]

    # Check the lengths
    if len(x_coords) < 1 or len(y_coords) < 1:
        raise ValueError(
            "Coordinate sequences need to be at least of length "
            "one or None (to use defaults), but were: {} and {}"
            "".format(x_coords, y_coords)
        )

    # Calculate distances, distinguishing between linear and logarithmic scale
    x_diffs, x_scale = determine_diffs_and_scale(
        x_coords, scale=x_scale, axis_name="x"
    )
    y_diffs, y_scale = determine_diffs_and_scale(
        y_coords, scale=y_scale, axis_name="y"
    )

    # Determine rectangle calculation method
    calc_rect = FUNC_MAP[(x_scale in LIN_SCALES, y_scale in LIN_SCALES)]

    # Need two additional borders: the lower border of the first rectangle and
    # the upper border of the last rectangle.
    x_diffs = expand_diffs(x_diffs, extend=extend, default=default_distance)
    y_diffs = expand_diffs(y_diffs, extend=extend, default=default_distance)

    # Divide by two to have the _sizes_ of the rectangle, not the distance
    # between the points, and then apply the size factor.
    x_sizes = x_diffs / 2.0 * size_factor
    y_sizes = y_diffs / 2.0 * size_factor

    # Create rects Dataset, which is populated below. It contains the positions
    # and sizes as separate data variables.
    shape = (len(x_coords), len(y_coords))
    rect_spec = np.dtype(
        [("xy", (float, (2,))), ("width", float), ("height", float)]
    )
    rects = xr.Dataset(
        data_vars=dict(
            pos_x=(("x", "y"), np.full(shape, np.nan)),
            pos_y=(("x", "y"), np.full(shape, np.nan)),
            len_x=(("x", "y"), np.full(shape, np.nan)),
            len_y=(("x", "y"), np.full(shape, np.nan)),
            rect_spec=(("x", "y"), np.full(shape, np.nan, dtype=rect_spec)),
        ),
        coords=dict(x=(("x",), x_coords), y=(("y",), y_coords)),
    )

    # Now, iterate over the coordinate combinations, calculate the rectangle
    # specification, and store it in the data array.
    it = itertools.product(
        enumerate(rects.coords["x"]), enumerate(rects.coords["y"])
    )
    for (x_idx, x_coord), (y_idx, y_coord) in it:
        # Get the relevant sizes calculated from the diffs and calculate the
        # rectangle specification from that
        pos_x, pos_y, len_x, len_y = calc_rect(
            x_coord,
            y_coord,
            (x_sizes[x_idx], x_sizes[x_idx + 1]),  # left  # right
            (y_sizes[y_idx], y_sizes[y_idx + 1]),  # bottom
        )  # top

        # Store in the corresponding data variable in the xr.Dataset
        rects["pos_x"][x_idx, y_idx] = pos_x
        rects["pos_y"][x_idx, y_idx] = pos_y
        rects["len_x"][x_idx, y_idx] = len_x
        rects["len_y"][x_idx, y_idx] = len_y
        rects["rect_spec"][x_idx, y_idx] = (pos_x, pos_y), len_x, len_y

    # Calculate limits of the space that is covered by the rectangles
    x_lims = determine_limits(coords=x_coords, sizes=x_sizes, scale=x_scale)
    y_lims = determine_limits(coords=y_coords, sizes=y_sizes, scale=y_scale)

    # Done.
    return rects, dict(x=x_lims, y=y_lims)



"""Tests the DataManager and involved functions and classes."""

import logging
import os
import uuid

import numpy as np
import pytest
from paramspace import ParamSpace
from pkg_resources import resource_filename

import utopya.eval.containers as udc
import utopya.eval.groups as udg
from utopya import DataManager, Multiverse
from utopya.eval.datamanager import _condense_thresh_func

from .. import ADVANCED_MODEL, DUMMY_MODEL, get_cfg_fpath

RUN_CFG_PATH = get_cfg_fpath("run_cfg.yml")
LARGE_SWEEP_CFG_PATH = get_cfg_fpath("large_sweep_cfg.yml")

# Suppress overly verbose log messages
logging.getLogger("utopya.task").setLevel(logging.DEBUG)
logging.getLogger("utopya.reporter").setLevel(logging.INFO)


# Fixtures --------------------------------------------------------------------
from .._fixtures import *


@pytest.fixture(autouse=True)
def register_demo_project(tmp_projects):
    """Use on all tests in this module"""
    pass


@pytest.fixture
def mv_kwargs(tmpdir) -> dict:
    """Returns a dict that can be passed to Multiverse for initialisation.

    This uses the `tmpdir` fixture provided by pytest, which creates a unique
    temporary directory that is removed after the tests ran through.
    """
    # Create a random string to use as model note
    rand_str = "test_" + uuid.uuid4().hex

    # Create a dict that specifies a unique testing path.
    return dict(
        model_name=ADVANCED_MODEL,
        run_cfg_path=RUN_CFG_PATH,
        user_cfg_path=False,  # to omit the user config
        paths=dict(out_dir=str(tmpdir), model_note=rand_str),
    )


@pytest.fixture
def dm_dummy_model(mv_kwargs, with_test_models) -> DataManager:
    """Initialises a Multiverse with a DataManager, runs a simulation with
    output going into a temporary directory, then returns the DataManager."""
    mv_kwargs["model_name"] = DUMMY_MODEL
    mv_kwargs["run_cfg_path"] = RUN_CFG_PATH
    mv = Multiverse(**mv_kwargs)
    mv.run_single()
    return mv.dm


@pytest.fixture
def dm_after_single(mv_kwargs, with_test_models) -> DataManager:
    """Initialises a Multiverse with a DataManager, runs a simulation with
    output going into a temporary directory, then returns the DataManager."""
    mv_kwargs["run_cfg_path"] = RUN_CFG_PATH
    mv = Multiverse(**mv_kwargs)
    mv.run_single()
    return mv.dm


@pytest.fixture
def dm_after_large_sweep(mv_kwargs, with_test_models) -> DataManager:
    """Initialises a Multiverse with a DataManager, runs a simulation with
    output going into a temporary directory, then returns the DataManager."""
    mv_kwargs["run_cfg_path"] = LARGE_SWEEP_CFG_PATH
    mv = Multiverse(**mv_kwargs)
    mv.run_sweep()
    return mv.dm


# Tests -----------------------------------------------------------------------


def test_condense_thresh_func():
    """Tests the function that evaluates the condense threshold for the
    data tree.
    """
    for l, n, t in zip([1, 4, 5], [42, 42, 42], [42, 42, 142]):
        _condense_thresh_func(level=l, num_items=n, total_item_count=t)


def test_init(tmpdir):
    """Tests simple initialisation of the data manager"""
    DataManager(str(tmpdir))


def test_simple_load(dm_dummy_model):
    """Tests the loading of simulation data for a single simulation"""
    dm = dm_dummy_model

    # Load and print a tree of the loaded data
    dm.load_from_cfg(print_tree=True)

    # Check that the config is loaded as expected
    assert "cfg" in dm
    assert "cfg/base" in dm
    assert "cfg/meta" in dm
    assert "cfg/model" in dm
    assert "cfg/run" in dm

    # Check that 'multiverse' is a MultiverseGroup
    assert "multiverse" in dm
    assert isinstance(dm["multiverse"], udg.MultiverseGroup)
    assert isinstance(dm["multiverse"].pspace, ParamSpace)

    assert len(dm["multiverse"]) == 1
    assert 0 in dm["multiverse"]
    uni = dm["multiverse"][0]

    # Check that the uni config is loaded
    assert "cfg" in uni

    # Binary data should NOT have been loaded by default because there is no
    # load configuration available for this model
    assert "data" not in uni


def test_load_single(dm_after_single):
    """Tests the loading of simulation data for a single simulation"""
    dm = dm_after_single

    # Load and print a tree of the loaded data
    dm.load_from_cfg(print_tree=True)

    # Check that the config is loaded as expected
    assert "cfg" in dm
    assert "cfg/base" in dm
    assert "cfg/meta" in dm
    assert "cfg/model" in dm
    assert "cfg/run" in dm

    # Check that 'multiverse' is a MultiverseGroup
    assert "multiverse" in dm
    assert isinstance(dm["multiverse"], udg.MultiverseGroup)
    assert isinstance(dm["multiverse"].pspace, ParamSpace)

    assert len(dm["multiverse"]) == 1
    assert 0 in dm["multiverse"]
    uni = dm["multiverse"][0]

    # Check that the uni config is loaded
    assert "cfg" in uni

    # Check that the binary data is loaded as expected
    assert "data" in uni
    assert f"data/{ADVANCED_MODEL}" in uni

    # Get the state dataset and check its content
    dset = uni[f"data/{ADVANCED_MODEL}/state"]
    print(dset.data)

    assert isinstance(dset, (udc.NumpyDC, udc.XarrayDC))
    assert dset.shape[1] == 100
    assert str(dset.dtype).startswith("f")


def test_load_sweep(dm_after_large_sweep):
    """Tests the loading of simulation data for a sweep"""
    dm = dm_after_large_sweep

    # Load and print a tree of the loaded data
    dm.load_from_cfg(print_tree="condensed")

    # Check that the config is loaded as expected
    assert "cfg" in dm
    assert "cfg/base" in dm
    assert "cfg/meta" in dm
    assert "cfg/model" in dm
    assert "cfg/run" in dm

    # Check that 'multiverse' is a MultiverseGroup of right length
    assert "multiverse" in dm
    assert isinstance(dm["multiverse"], udg.MultiverseGroup)
    assert isinstance(dm["multiverse"].pspace, ParamSpace)
    assert 0 not in dm["multiverse"]
    assert len(dm["multiverse"]) == dm["multiverse"].pspace.volume

    # Now go over all available universes
    for uni_no, uni in dm["multiverse"].items():
        # Check that the uni config is loaded
        assert "cfg" in uni

        # Check that the binary data is loaded as expected
        assert "data" in uni
        assert f"data/{ADVANCED_MODEL}" in uni
        assert f"data/{ADVANCED_MODEL}/state" in uni

        # Get the state dataset and check its content
        dset = uni[f"data/{ADVANCED_MODEL}/state"]

        assert isinstance(dset, (udc.NumpyDC, udc.XarrayDC))
        assert dset.shape == (uni["cfg"]["num_steps"] + 1, 100)
        assert str(dset.dtype).startswith("f")



"""Implements generally useful functions, partly by importing from dantro"""

import io
import logging
import os
import re
import subprocess
import sys
from datetime import timedelta
from typing import Any, Callable, Sequence, Tuple, Union

from dantro.tools import (
    IS_A_TTY,
    TTY_COLS,
    center_in_line,
    fill_line,
    format_time,
    make_columns,
    print_line,
    recursive_getitem,
    recursive_update,
)

from ._yaml import load_yml, write_yml, yaml

log = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
# working on dicts ------------------------------------------------------------


def load_selected_keys(
    src: dict,
    *,
    add_to: dict,
    keys: Sequence[Tuple[str, type, bool]],
    err_msg_prefix: str = None,
    prohibit_unexpected: bool = True,
) -> None:
    """Loads (only) selected keys from dict ``src`` into dict ``add_to``.

    Args:
        src (dict): The dict to load values from
        add_to (dict): The dict to load values into
        keys (Sequence[Tuple[str, type, bool]]): Which keys to load, given
            as sequence of ``(key, allowed types, [required=False])`` tuples.
        err_msg_prefix (str): A description string, used in error message
        prohibit_unexpected (bool, optional): Whether to raise on keys
            that were unexpected, i.e. not given in ``keys`` argument.

    Raises:
        KeyError: On missing key in ``src``
        TypeError: On bad type of value in ``src``
        ValueError: On unexpected keys in ``src``
    """

    def unpack(spec) -> Tuple[str, Union[type, Sequence[type]], bool]:
        """Unpacks a schema entry into a 3-tuple"""
        if len(spec) == 3:
            k, allowed_types, required = spec
        else:
            k, allowed_types = spec
            required = False

        return k, allowed_types, required

    for spec in keys:
        k, allowed_types, required = unpack(spec)
        if k not in src:
            if not required:
                continue
            raise ValueError(
                "{}Missing required key: {}"
                "".format(err_msg_prefix + " " if err_msg_prefix else "", k)
            )

        if not isinstance(src[k], allowed_types):
            raise TypeError(
                "{}Bad type for value of '{}'! Expected "
                "{} but got {}: {}"
                "".format(
                    err_msg_prefix + " " if err_msg_prefix else "",
                    k,
                    allowed_types,
                    type(src[k]),
                    src[k],
                )
            )

        add_to[k] = src[k]

    if not prohibit_unexpected:
        return

    allowed_keys = tuple(unpack(spec)[0] for spec in keys)
    unexpected_keys = tuple(k for k in src if k not in allowed_keys)
    if unexpected_keys:
        raise ValueError(
            "{}Received unexpected keys: {}\n"
            "Expected only: {}"
            "".format(
                err_msg_prefix + " " if err_msg_prefix else "",
                ", ".join(unexpected_keys),
                ", ".join(allowed_keys),
            )
        )


def add_item(
    value,
    *,
    add_to: dict,
    key_path: Sequence[str],
    value_func: Callable = None,
    is_valid: Callable = None,
    ErrorMsg: Callable = None,
) -> None:
    """Adds the given value to the ``add_to`` dict, traversing the given key
    path. This operation happens in-place.

    Args:
        value: The value of what is to be stored. If this is a callable, the
            result of the call is stored.
        add_to (dict): The dict to add the entry to
        key_path (Sequence[str]): The path at which to add it
        value_func (Callable, optional): If given, calls it with ``value`` as
            argument and uses the return value to add to the dict
        is_valid (Callable, optional): Used to determine whether ``value`` is
            valid or not; should take single positional argument, return bool
        ErrorMsg (Callable, optional): A raisable object that prints an error
            message; gets passed ``value`` as positional argument.

    Raises:
        Exception: type depends on specified ``ErrorMsg`` callable
    """
    # Check the value by calling the function; it should raise an error
    if is_valid is not None:
        if not is_valid(value):
            raise ErrorMsg(value)

    # Determine which keys need to be traversed
    traverse_keys, last_key = key_path[:-1], key_path[-1]

    # Set the starting point
    d = add_to

    # Traverse keys
    for _key in traverse_keys:
        # Check if a new entry is needed
        if _key not in d:
            d[_key] = dict()

        # Select the new entry
        d = d[_key]

    # Now d is where the value should be added
    # If applicable
    if value_func is not None:
        value = value_func(value)

    # Store in dict, mutable. Done.
    d[last_key] = value


# -- String formatting --------------------------------------------------------


def pprint(obj: Any, **kwargs):
    """Prints a "pretty" string representation of the given object.

    Args:
        obj (Any): The object to print
        **kwargs: Passed to print
    """
    print(pformat(obj), **kwargs)


def pformat(obj: Any) -> str:
    """Creates a "pretty" string representation of the given object.

    This is achieved by creating a yaml representation.

    .. todo::

        Improve parsing of leaf-level mappings
    """
    sstream = io.StringIO("")
    yaml.dump(obj, stream=sstream)
    sstream.seek(0)
    return sstream.read()


# misc ------------------------------------------------------------------------


def parse_si_multiplier(s: str) -> int:
    """Parses a string like ``1.23M`` or ``-2.34 k`` into an integer.

    If it is a string, parses the SI multiplier and returns the appropriate
    integer for use as number of simulation steps.
    Supported multipliers are ``k``, ``M``, ``G`` and ``T``. These need to be
    used as the suffix of the string.

    .. note::

        This is only intended to be used with integer values and does *not*
        support float values like ``1u``.

    The used regex can be found `here <https://regex101.com/r/xngAoc/1>`_.

    Args:
        s (str): A string representing an integer number, potentially including
            a supported SI multiplier as *suffix*.

    Returns:
        int: The parsed number of steps as integer. If the value has decimal
            places, integer rounding is applied.

    Raises:
        ValueError: Upon string that does not match the expected pattern
    """
    SUFFIXES = dict(k=1e3, M=1e6, G=1e9, T=1e12)
    pattern = r"^(?P<value>\-?\s?\d+|\-?\s?\d+\.\d+)?\s?(?P<suffix>[kMGT])?$"
    # See:  https://regex101.com/r/xngAoc/1

    match = re.match(pattern, s.strip())
    if not match:
        raise ValueError(
            f"Cannot parse '{s}' into an integer! May only contain the metric "
            "suffixes k, M, G, or T. Examples: 1000, 1k, 1.23M, 0.5 G"
        )

    groups = match.groupdict()
    val = float(groups["value"].replace(" ", ""))
    mul = SUFFIXES[groups["suffix"]] if groups["suffix"] else 1

    return int(val * mul)


def parse_num_steps(
    N: Union[str, int], *, raise_if_negative: bool = True
) -> int:
    """Given a string like ``1.23M`` or an integer, prepares the num_steps
    argument for a single universe simulation.

    For string arguments, uses :py:func:`~utopya.tools.parse_si_multiplier` for
    string parsing. If that fails, attempts to read it in float notation by
    calling ``int(float(N))``.

    .. note:: This function always applies integer rounding.

    Args:
        N (Union[str, int]): The ``num_steps`` argument as a string or integer.
        raise_if_negative (bool, optional): Whether to raise an error if the
            value is negative.

    Returns:
        int: The parsed value for ``num_steps``

    Raises:
        ValueError: Result invalid, i.e. not parseable or of negative value.
    """
    if isinstance(N, str):
        try:
            N = parse_si_multiplier(N)

        except ValueError as err:
            # Don't give up just yet, could still be in scientific notation ...
            try:
                N = int(float(N))

            except:
                # Ok, that also failed. Giving up now.
                raise ValueError(f"Failed parsing `num_steps`: {err}") from err

    if raise_if_negative and N < 0:
        raise ValueError(
            f"Argument `num_steps` needs to be non-negative, but was {N}!"
        )

    return N



"""Defines the utopya CLI"""

import sys

import click

from .batch import batch
from .config import config
from .eval import evaluate
from .models import models
from .projects import projects
from .run import run

cli = click.Group(
    help=(
        "utopya: a versatile simulation runner and manager\n\n"
        "utopya's main feature is to configure, run, and evaluate computer "
        "simulations. "
        "In the minimalistic use case, utopya is a simple runner for an "
        "arbitrary executable. "
        "If complying with the full utopya interface, it provides a highly "
        "configurable feature set, including a hierarchical configuration "
        "structure, simulation monitoring, and coupling to an automated "
        "data processing pipeline for simulation output."
    ),
    epilog=(
        "Copyright (C) 2018 – 2022  utopya developers\n\n"
        "utopya is free software and comes with absolutely no warranty. "
        "You are welcome to redistribute it under the conditions specified in "
        "the LGPLv3+ license. "
        "For more information, visit:\n\n"
        "   utopia-project.org  |  gitlab.com/utopia-project/utopya"
    ),
)

cli.add_command(run)
cli.add_command(evaluate)
cli.add_command(batch)
cli.add_command(config)
cli.add_command(models)
cli.add_command(projects)



"""Helper module that contains tools useful for module imports or manipulation
of the system path.
These are not implemented here but in :py:mod:`dantro._import_tools`.
"""

from dantro._import_tools import *



"""Implements the utopya run CLI subtree"""

import click

from ._shared import OPTIONS, add_options, default_none
from ._utils import Echo


@click.command(
    help=(
        "Perform a simulation run and evaluate its outputs.\n"
        "\n"
        "Starts a simulation of the MODEL_NAME model, allowing to pass "
        "a custom RUN_CFG and otherwise manipulating the meta configuration. "
        "Subsequently, the simulation output is fed into the evaluation "
        "pipeline."
    ),
)
@click.argument("model_name")
@click.argument(
    "run_cfg",
    required=False,
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
)
@add_options(OPTIONS["model_selection"])  # --label, --cfg-set
#
# -- Update the meta configuration
#
@click.option(
    "--validate/--no-validate",
    "validate",
    default=None,
    help=(
        "If given, sets whether there will be parameter validation. "
        "If not given, will use the default specified in the meta config."
    ),
)
@click.option(
    "--run-mode",
    default=None,
    type=click.Choice(("single", "sweep")),
    help=(
        "Forces a single simulation or a sweep. If not given, will use the "
        "value specified in the meta configuration."
    ),
)
@click.option(
    "--note",
    default=None,
    type=str,
    help=(
        "A suffix that is appended to the name of the run output directory; "
        "this can be useful to give a name to a certain simulation run."
    ),
)
@click.option(
    "-N",
    "--num-steps",
    default=None,
    type=str,
    help=(
        "Sets the number of simulation steps. Needs to be an integer. Metric "
        "suffixes (k, M, G, T) can be used to denote large numbers, e.g "
        "`1.23M` translating to `1230000`, and scientific notation is also "
        "supported (applying integer rounding)."
    ),
)
@click.option(
    "--write-every",
    "--we",
    default=None,
    type=str,
    help=(
        "Sets the `write_every` parameter, controlling how frequently model "
        "data is written. "
        "Can be given in the same formats as --num-steps."
    ),
)
@click.option(
    "--write-start",
    "--ws",
    default=None,
    type=str,
    help=(
        "Sets the `write_start` parameter, specifying the first time step "
        "at which data is written. After that step, data is written every "
        "`write_every`th step."
        "Can be given in the same formats as --num-steps."
    ),
)
@click.option(
    "--num-seeds",
    default=None,
    type=click.IntRange(min=1),
    help=(
        "Creates a parameter dimension for the seeds with the given number "
        "of seed values. This also sets the `perform_sweep` parameter to "
        "True, such that a sweep is invoked."
    ),
)
@click.option(
    "--set-model-params",
    "--mp",
    multiple=True,
    callback=default_none,
    help=(
        "Sets entries in the model configuration, i.e. inside of the "
        "`parameter_space.<model_name>` entry of the meta configuration. "
        "Example: --mp some.param=42 sets the 'param' entry in 'some' to 42. "
        "Specify DELETE as value to remove an entry. "
        "Repeat the --mp option to set multiple values."
    ),
)
@click.option(
    "--set-pspace-params",
    "--pp",
    multiple=True,
    callback=default_none,
    help=(
        "Like --set-model-params but attaching to the `parameter_space` entry "
        "of the meta configuration. "
        "Repeat the --pp option to set multiple values."
        "These arguments are parsed after --set-pspace-params and "
        "--set-model-params such that they can overwrite any of the "
        "previously defined arguments."
    ),
)
@click.option(
    "--set-params",
    "-p",
    multiple=True,
    callback=default_none,
    help=(
        "Like --set-model-params but attaching to the root level of the meta "
        "configuration. "
        "These arguments are parsed after --set-pspace-params and "
        "--set-model-params such that they can overwrite any of the "
        "previously defined arguments."
    ),
)
#
# -- Evaluation
#
@click.option(
    "--eval/--no-eval",
    "perform_eval",
    default=None,
    help=(
        "If given, overwrites the default behavior of whether the simulation "
        "run should be followed by the evaluation routine (data loading "
        "and plotting) or not. "
        "The default can also be set in the model info file. "
        "If no default is given there, will attempt evaluation."
    ),
)
@add_options(OPTIONS["load"])
@add_options(OPTIONS["eval"])
@add_options(OPTIONS["debug_flag"])  # --debug
@add_options(OPTIONS["cluster_mode"])  # --cluster-mode
#
#
#
@click.pass_context
def run(ctx, **kwargs):
    """Invokes a model simulation run and subsequent evaluation"""
    import utopya
    from utopya.exceptions import ValidationError
    from utopya.tools import pformat

    from ._utils import parse_run_and_plots_cfg, parse_update_dicts
    from .eval import _load_and_eval

    _log = utopya._getLogger("utopya")  # TODO How best to do this?!

    # Preparations . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
    _log.info("Parsing additional command line arguments ...")
    update_dict, update_plots_cfg = parse_update_dicts(
        _mode="run", **kwargs, _log=_log
    )

    if update_dict:
        _log.note("Updates to meta configuration:\n\n%s", pformat(update_dict))

    model = utopya.Model(
        name=kwargs["model_name"],
        bundle_label=kwargs["label"],
    )

    run_cfg, plots_cfg = parse_run_and_plots_cfg(
        model=model,
        run_cfg=kwargs["run_cfg"],
        plots_cfg=kwargs["plots_cfg"],
        cfg_set=kwargs["cfg_set"],
        _log=_log,
    )
    kwargs["plots_cfg"] = plots_cfg
    kwargs["update_plots_cfg"] = update_plots_cfg

    mv = model.create_mv(run_cfg_path=run_cfg, **update_dict)

    # Running the simulation . . . . . . . . . . . . . . . . . . . . . . . . .
    mv.run()

    # Check whether to start evaluation routine
    perform_eval = model.info_bundle.eval_after_run
    _perform_eval = kwargs.pop("perform_eval")
    if _perform_eval is not None:
        perform_eval = _perform_eval

    if perform_eval is False:
        _log.note("Evaluation routine was not enabled for this model.")
        _log.remark("Use --eval / --no-eval to overwrite model defaults.")
        _log.progress("Exiting now ...\n")
        return

    # Loading and evaluating . . . . . . . . . . . . . . . . . . . . . . . . .
    _load_and_eval(
        _log=_log,
        ctx=ctx,
        mv=mv,
        **kwargs,
    )



"""Implements the `utopya models` subcommand tree of the CLI"""

import sys
from typing import Tuple

import click

from ._shared import OPTIONS, add_options
from ._utils import Echo

models = click.Group(
    name="models",
    help="Show available models and register new ones.",
)


# -- Locally relevant utility functions ---------------------------------------


def _evaluate_fstr_for_list(*, fstr: str, model_names: str, sep: str) -> str:
    """Evaluates a format string using the information from a list of model
    names.

    Args:
        fstr (str): The format string to evaluate for each model name
        model_names (str): A splittable string of model names
        sep (str): The separator used to split the ``model_names`` string
    """
    return sep.join(fstr.format(model_name=m) for m in model_names.split(sep))


# -- Utility commands ---------------------------------------------------------
# .. utopya models ls .........................................................


@models.command(
    name="ls",
    help="Lists all registered models",
)
@click.option(
    "-l",
    "--long",
    "long_mode",
    is_flag=True,
    help="Show more detailed information.",
)
def list_models(long_mode: bool):
    import utopya

    if long_mode:
        click.echo(utopya.MODELS.info_str_detailed)
    else:
        click.echo(utopya.MODELS.info_str)


# .. utopya models rm .........................................................


@models.command(
    name="rm",
    help="Removes info bundles from individual models",
)
@click.argument("model_name")
@add_options(OPTIONS["label"])
@click.option(
    "-a",
    "--all",
    "remove_all",
    is_flag=True,
    help="If set, removes all info bundles and deletes the registry entry",
)
@click.option(
    "-y",
    "--yes",
    "removal_confirmed",
    is_flag=True,
    help="If set, skips the confirmation prompt for removing all info bundles",
)
def remove_model_or_bundle(
    *, model_name: str, label: str, remove_all: bool, removal_confirmed: bool
):
    import utopya

    if not remove_all:
        if not label:
            _avail = ", ".join(utopya.MODELS[model_name].keys())
            Echo.info(f"Available bundles for model '{model_name}':  {_avail}")
            label = click.prompt("Which bundle would you like to remove?")

        Echo.info(
            f"Removing info bundle '{label}' for model '{model_name}'..."
        )
        utopya.MODELS[model_name].pop(label)
        Echo.success(
            f"Info bundle labelled '{label}' removed "
            f"from registry entry of model '{model_name}'."
        )

    else:
        Echo.info(f"Removing registry entry for model '{model_name}'...")
        if not removal_confirmed and not click.confirm("Are you sure?"):
            Echo.info("Not removing anything.")
            sys.exit(0)

        utopya.MODELS.remove_entry(model_name)
        Echo.success(f"Registry entry for model '{model_name}' removed.")


# .. utopya models edit .......................................................


@models.command(help="Edit the model registry entry")
@click.argument("model_name")
def edit(*, model_name: str):
    """Edits the model registry entry of the given model"""
    Echo.info(f"Opening '{model_name}' model's registry file for editing ...")
    import utopya

    Echo.warning("Take care not to corrupt the file!")
    if not click.confirm("Open file for editing?"):
        Echo.info("Not opening.")
        sys.exit(0)

    try:
        click.edit(
            filename=utopya.MODELS[model_name].registry_file_path,
            extension=".yml",
        )

    except Exception as exc:
        Echo.error("Editing model registry file failed!", error=exc)
        sys.exit(1)

    Echo.success(f"Successfully edited registry file of '{model_name}' model.")


# .. utopya models set-default ................................................


@models.command(help="Sets the default info bundle to use for a model")
@click.argument("model_name")
@click.argument("label")
def set_default(*, model_name: str, label: str):
    """Sets the default info bundle to use for a certain model"""
    Echo.info(
        f"Setting the info bundle labelled '{label}' as the default "
        f"for '{model_name}' ..."
    )

    import utopya

    utopya.MODELS[model_name].default_label = label
    Echo.success(f"Successully set default label for model '{model_name}'.")


# .. utopya models info .......................................................
# TODO Expand to show more information


@models.command(
    help=(
        "Shows model information.\n"
        "\n"
        "Currently, this only shows the available configuration set names."
    )
)
@click.argument("model_name")
@add_options(OPTIONS["label"])
def info(*, model_name: str, label: str):
    import utopya
    from utopya.tools import make_columns

    _log = utopya._getLogger("utopya_cli")

    model = utopya.Model(name=model_name, bundle_label=label)

    _log.info("Fetching available config sets ...")
    cfg_sets = model.default_config_sets
    if cfg_sets:
        _log.note(
            "Have %d config sets available for model '%s':\n%s",
            len(cfg_sets),
            model.name,
            make_columns(cfg_sets),
        )

    else:
        _log.note(
            "There are no config sets available for model '%s'.", model.name
        )

    _log.remark(
        "To add config sets, create subdirectories containing run.yml and/or "
        "eval.yml files in one of the search directories listed above.",
    )


# .. utopya models copy .......................................................
# TODO Implement


@models.command(
    help=("Copies a model implementation.\n\nThis is not implemented yet!")
)
@click.argument("model_name")
@add_options(OPTIONS["label"])
@click.option(
    "--new-name",
    prompt=True,
    help="Name of the new model. If not given, will prompt for it.",
)
@click.option(
    "--target-project",
    prompt=True,
    help=(
        "Name of the utopya project to copy the new model to. If not given, "
        "will prompt for it. "
        "Note that this project needs to be known by utopya; "
        "use `utopya projects register` to register it first."
    ),
)
@click.option(
    "--dry-run",
    flag_value=True,
    help=(
        "Perform a dry run: No copy or write operations will be carried out."
    ),
)
@click.option(
    "--skip-exts",
    show_default=True,
    default="pyc",
    callback=lambda c, _, val: val.split(),
    help=(
        "File extensions to skip. "
        "To pass multiple values, use quotes and separate individual "
        "extensions using spaces. There should not be any leading dots!"
    ),
)
@click.option(
    "--register",
    flag_value=True,
    help=("If given, will also register the newly created model."),
)
def copy(
    *,
    model_name: str,
    label: str,
    new_name: str,
    target_project: str,
    dry_run: bool,
    skip_exts: str,
    register: bool,
):
    raise NotImplementedError("copy")  # TODO


# -- Registration -------------------------------------------------------------
# .. utopya models register ...................................................

register = click.Group(
    name="register",
    help="Register new models, either individually or in a batch.",
)
models.add_command(register)

# .. utopya models register single ............................................


@register.command(
    name="single",
    help="Register a single new model or update an existing one.",
)
@click.argument("model_name")
@click.option(
    "-e",
    "--executable",
    required=True,
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help="Path to the model executable.",
)
@click.option(
    "--source-dir",
    default=None,
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, resolve_path=True
    ),
    help=(
        "Path to the directory that contains the model's source files; "
        "this can be used to automatically extract model-related information "
        "like the default model configuration or plot-related configurations."
    ),
)
@click.option(
    "--default-cfg",
    default=None,
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help="Path to the model's default configuration file.",
)
@click.option(
    "--plots-cfg",
    default=None,
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help="Path to the model's default plots configuration file.",
)
@click.option(
    "--label",
    type=click.STRING,
    default="set_via_cli",
    help=(
        "Label to identify the info bundle with; if not given, will use a "
        "default value. This allows registering multiple versions of a model "
        "under the same name."
    ),
)
@click.option(
    "--set-default",
    "set_as_default",
    is_flag=True,
    default=None,
    help=("Whether to set the registered model(s) as default."),
)
@click.option(
    "--project-name",
    type=click.STRING,
    default=None,
    help="Name of the project this model is part of.",
)
@click.option(
    "--exists-action",
    default="validate",
    type=click.Choice(("skip", "raise", "validate", "overwrite")),
    help=(
        "Which action to take upon an existing bundle with the same label. "
        "By default, validates the to-be-added information against a "
        "potentially existing bundle."
    ),
)
def register_single(
    *,
    model_name: str,
    executable: str,
    source_dir: str,
    default_cfg: str,
    plots_cfg: str,
    label: str,
    set_as_default: bool,
    project_name: str,
    exists_action: str,
):
    """Registers a new model"""
    Echo.progress(f"Registering model '{model_name}' (label: '{label}')...")
    Echo.remark(f"  executable:        {executable}")
    Echo.remark(f"  source directory:  {source_dir}")
    Echo.remark(f"  default config:    {default_cfg}")
    Echo.remark(f"  plots config:      {plots_cfg}")

    bundle_kwargs = dict(
        label=label,
        paths=dict(
            executable=executable,
            default_cfg=default_cfg,
            source_dir=source_dir,
            plots_cfg=plots_cfg,
        ),
        project_name=project_name,
    )

    import utopya

    try:
        utopya.MODELS.register_model_info(
            model_name,
            exists_action=exists_action,
            set_as_default=set_as_default,
            extract_model_info=True,
            **bundle_kwargs,
        )

    except Exception as exc:
        Echo.error("Registration failed!", error=exc)
        sys.exit(1)

    else:
        Echo.success(f"Successully registered model '{model_name}'.")


# .. utopya models register batch .............................................


@register.command(
    name="from-list",
    help=(
        "Register multiple models by providing aligned lists of their names, "
        "the paths to their executables and their source directories. "
        "Any additional information is attempted to be extracted from the "
        "corresponding model source directory or a potentially existing model "
        "information file.\n"
        "Note that the arguments need to be separable lists, where the "
        "`--separator` argument determines the separation string. "
        "Also, arguments probably need to be put into quotes and spaces need "
        "to be escaped."
    ),
)
@click.argument("model_names", type=click.STRING)
@click.option(
    "--executables",
    type=click.STRING,
    help=(
        "A list of paths pointing to the model executables. "
        "Paths may be given as relative to the --base-executable-dir. "
        "If all paths match a pattern, consider using the "
        "--executable-fstr argument instead. "
        "One of these arguments *needs* to be given."
    ),
)
@click.option(
    "--source-dirs",
    type=click.STRING,
    help=(
        "A list of paths pointing to the model source directories. "
        "Paths may be given as relative to the --base-source-dir. "
        "If all paths match a pattern, consider using the "
        "--source-dir-fstr argument instead."
    ),
)
@click.option(
    "--base-executable-dir",
    type=click.Path(file_okay=False, exists=True, resolve_path=True),
    help="If given, relative executable paths are interpreted against this.",
)
@click.option(
    "--base-source-dir",
    type=click.Path(file_okay=False, exists=True, resolve_path=True),
    help=(
        "If given, relative source directory paths are interpreted against "
        "this."
    ),
)
@click.option(
    "--executable-fstr",
    type=click.STRING,
    help=(
        "A format string that can be used instead of the --executables "
        "argument and is evaluated for each entry in MODEL_NAMES."
    ),
)
@click.option(
    "--source-dir-fstr",
    type=click.STRING,
    help=(
        "A format string that can be used instead of the --source-dirs "
        "argument and is evaluated for each entry in MODEL_NAMES."
    ),
)
@click.option(
    "--py-tests-dir-fstr",
    type=click.STRING,
    help=(
        "A format string that can be used to set the models python tests "
        "directory."
    ),
)
@click.option(
    "--py-plots-dir-fstr",
    type=click.STRING,
    help=(
        "A format string that can be used to set the models python tests "
        "directory."
    ),
)
@click.option(
    "--separator",
    default=";",
    help=(
        "By which separator to split the --model-names, --executables, and "
        "--source-dirs arguments. Default: ';'"
    ),
)
@click.option(
    "--label",
    type=click.STRING,
    default="set_via_cli",
    help=(
        "Label to identify the info bundles with; if not given, will use a "
        "default value. This allows registering multiple versions of a model "
        "under the same name."
    ),
)
@click.option(
    "--set-default",
    "set_as_default",
    is_flag=True,
    default=None,
    help=("Whether to set the registered model(s) as default."),
)
@click.option(
    "--project-name",
    type=click.STRING,
    default=None,
    help="Name of the project these models are part of.",
)
@click.option(
    "--exists-action",
    default="validate",
    type=click.Choice(("skip", "raise", "validate", "overwrite")),
    help=(
        "Which action to take upon an existing bundle with the same label. "
        "By default, validates the to-be-added information with a potentially "
        "existing bundle; this will fail if the to-be-added bundle does not "
        "compare equal to an existing bundle with the same label."
    ),
)
def register_from_list(
    *,
    model_names: str,
    label: str,
    executables: str,
    source_dirs: str,
    executable_fstr: str,
    source_dir_fstr: str,
    separator: str,
    set_as_default: bool,
    exists_action: str,
    **more_paths,
):
    if executable_fstr:
        if executables:
            Echo.error(
                "Arguments --executable-fstr and --executables are mutually "
                "exclusive! Make sure to only pass one of them."
            )
            sys.exit(1)

        executables = _evaluate_fstr_for_list(
            model_names=model_names, fstr=executable_fstr, sep=separator
        )

    elif not executables:
        Echo.error("Missing argument --executables or --executable-fstr!")
        sys.exit(1)

    if source_dir_fstr:
        if source_dirs:
            Echo.error(
                "Arguments --source-dir-fstr and --source-dirs are mutually "
                "exclusive! Make sure to only pass one of them."
            )
            sys.exit(1)

        source_dirs = _evaluate_fstr_for_list(
            model_names=model_names, fstr=source_dir_fstr, sep=separator
        )

    # Everything ok, can start registering
    import utopya
    from utopya.model_registry._registration import register_models_from_list

    try:
        register_models_from_list(
            registry=utopya.MODELS,
            model_names=model_names,
            label=label,
            executables=executables,
            source_dirs=source_dirs,
            separator=separator,
            more_paths=more_paths,
            set_as_default=set_as_default,
            extract_model_info=True,
            exists_action=exists_action,
            _log=Echo,
        )

    except Exception as exc:
        Echo.error("Registration failed!", error=exc)
        raise
        sys.exit(1)


# .. utopya models register from-manifest .....................................


@register.command(
    name="from-manifest",
    help=(
        "Register one or many models using manifest files: YAML files that "
        "contains all relevant model information; see documentation for a "
        "valid manifest file format."
    ),
)
@click.argument(
    "manifest_files",
    nargs=-1,
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
)
@click.option(
    "--model-name",
    "custom_model_name",
    type=click.STRING,
    default=None,
    help=(
        "If only a single manifest file is given, this option can be used to "
        "specify the model name, ignoring the one given in the manifest file."
    ),
)
@click.option(
    "--label",
    "custom_label",
    type=click.STRING,
    default=None,
    help=(
        "If given, this label will be used instead of the one given in the "
        "manifest file(s). "
        "If no custom label is given and the manifest file does not define "
        "one either, the default will be 'from_manifest_file'."
    ),
)
@click.option(
    "--set-default",
    "set_as_default",
    is_flag=True,
    default=None,
    help=("Whether to set the registered model(s) as default."),
)
@click.option(
    "--project-name",
    "custom_project_name",
    type=click.STRING,
    default=None,
    help=(
        "This option can be used to specify the project name, ignoring the "
        "name(s) specified in the manifest file(s)."
    ),
)
@click.option(
    "--exists-action",
    default="validate",
    type=click.Choice(("skip", "raise", "validate", "overwrite")),
    help=(
        "Which action to take upon an existing bundle with the same label. "
        "By default, validates the to-be-added information with a potentially "
        "existing bundle; this will fail if the to-be-added bundle does not "
        "compare equal to an existing bundle with the same label."
    ),
)
def register_from_manifest(
    *,
    manifest_files: Tuple[str],
    custom_model_name: str,
    set_as_default: bool,
    custom_project_name: str,
    custom_label: str,
    exists_action: str,
):
    """Registers one or many models using manifest files"""
    num_files = len(manifest_files)

    if custom_model_name and num_files != 1:
        Echo.error(
            "A model name can only be specified if only a single "
            f"manifest file is given (got {num_files}). "
        )
        Echo.info(
            "Either remove the `--model-name` argument or make sure to pass "
            "only a single manifest file."
        )
        sys.exit(1)

    # All checks done, let's go
    Echo.info(f"Parsing information from {num_files} manifest file(s) ...")
    import utopya

    for i, manifest_file in enumerate(manifest_files):
        Echo.progress(
            f"\nRegistering model from manifest file {i + 1} / {num_files} ..."
        )
        Echo.remark(f"File:  {manifest_file}")

        bundle_kwargs = utopya.tools.load_yml(manifest_file)

        # Handle custom model name, label, or project name
        model_name = bundle_kwargs.pop("model_name")
        if custom_model_name:
            Echo.note(f"Using custom model name:     {custom_model_name}")
            model_name = custom_model_name
        else:
            Echo.note(f"Model name:                  {model_name}")

        label = bundle_kwargs.pop("label", "from_manifest_file")
        if custom_label:
            Echo.note(f"Using custom label:          {custom_label}")
            label = custom_label
        else:
            Echo.note(f"Label:                       {label}")

        Echo.note(f"Setting as default?          {set_as_default}")

        project_name = bundle_kwargs.pop("project_name", None)
        if custom_project_name:
            Echo.note(f"Using custom project name:   {custom_project_name}")
            project_name = custom_project_name

        # Also add path to manifest file in the paths dict, such that the
        # info bundle knows about it. Then register.
        utopya.tools.add_item(
            manifest_file,
            add_to=bundle_kwargs,
            key_path=("paths", "model_info"),
        )
        try:
            utopya.MODELS.register_model_info(
                model_name,
                label=label,
                project_name=project_name,
                exists_action=exists_action,
                extract_model_info=False,  # already done above
                set_as_default=set_as_default,
                **bundle_kwargs,
            )

        except Exception as exc:
            Echo.error("Registration failed!", error=exc)
            sys.exit(1)

        Echo.info(
            f"Successfully registered model information for '{model_name}', "
            f"labelled '{label}':"
        )
        Echo.remark(utopya.tools.pformat(utopya.MODELS[model_name][label]))

    Echo.success(f"Model information registered successully.")



"""Implements data container classes specialised on Utopia output data.

It is based on the :py:class:`dantro.base.BaseDataContainer` classes,
especially their numeric form, the
:py:class:`~dantro.containers.numeric.NumpyDataContainer` and the
:py:class:`~dantro.containers.xr.XrDataContainer`.
"""

import logging
from typing import Sequence, Tuple, Union

import dantro.containers
import dantro.mixins
import numpy as np
import xarray as xr

from ..tools import yaml as _yaml

log = logging.getLogger(__name__)


# -----------------------------------------------------------------------------


class NumpyDC(
    dantro.mixins.Hdf5ProxySupportMixin, dantro.containers.NumpyDataContainer
):
    """This is the base class for numpy data containers used in Utopia.

    It is based on the NumpyDataContainer provided by dantro and extends it
    with the Hdf5ProxySupportMixin, allowing to load the data from the Hdf5
    file only once it becomes necessary.
    """


class XarrayDC(
    dantro.mixins.Hdf5ProxySupportMixin, dantro.containers.XrDataContainer
):
    """This is the base class for xarray data containers used in Utopia.

    It is based on the XrDataContainer provided by dantro. As of now, it has
    no proxy support, but will gain it once available on dantro side.
    """

    # Configure proxy support .................................................
    PROXY_RESOLVE_ASTYPE = None
    """Which type to resolve the proxy to.
    None defaults to :py:class:`numpy.ndarray`."""

    PROXY_RETAIN = True
    """Whether to retain a proxy after resolving it; allows reinstating
    proxy objects."""

    PROXY_REINSTATE_FAIL_ACTION = "log_warning"
    """Which action to take if reinstating a proxy was not possible"""

    # Specialize XrDataContainer for Utopia ...................................
    _XRC_DIMS_ATTR = "dim_names"
    """Define as class variable the name of the attribute that determines the
    dimensions of the :py:class:`xarray.DataArray`.
    """

    _XRC_DIM_NAME_PREFIX = "dim_name__"
    """Attributes prefixed with this string can be used to set names for
    specific dimensions. The prefix should be followed by an integer-parsable
    string, e.g. ``dim_name__0`` would be the dimension name for the 0th dim
    """

    _XRC_COORDS_ATTR_PREFIX = "coords__"
    """Attributes prefixed with this string determine the coordinate values for
    a specific dimension. The prefix should be followed by the _name_ of the
    dimension, e.g. ``coords__time``. The values are interpreted according to
    the default coordinate mode or, if given, the ``coords_mode__*`` attribute
    """

    _XRC_COORDS_MODE_DEFAULT = "values"
    """The default mode by which coordinates are interpreted. See the base
    class, :py:class:`dantro.containers.xr.XrDataContainer` for more
    information.

    Available modes:
      - ``values``: the explicit values (iterable) to use for coordinates
      - ``trivial``: The trivial indices; ignores the coordinate argument
      - ``scalar``: makes sure only a single coordinate is provided
      - ``range``: python built-in range function arguments
      - ``arange``: np.arange arguments
      - ``linspace``: np.linspace arguments
      - ``logspace``: np.logspace arguments
      - ``start_and_step``: the start and step values of an integer range
          expression; the stop value is deduced by looking at the length
          of the corresponding dimension.
      - ``linked``: Load the coordinates from a linked object within the
          tree, specified by a relative path from the current object.
    """

    _XRC_COORDS_MODE_ATTR_PREFIX = "coords_mode__"
    """Prefix for the coordinate mode if a custom mode is to be used.
    To, e.g., use mode ``start_and_step`` for time dimension, set the
    ``coords_mode__time`` attribute to value ``start_and_step``
    """

    _XRC_INHERIT_CONTAINER_ATTRIBUTES = False
    """Whether to inherit the other container attributes"""

    _XRC_STRICT_ATTR_CHECKING = True
    """Whether to use strict attribute checking; throws errors if there are
    container attributes available that match the prefix but don't match a
    valid dimension name. Can be disabled for speed improvements.
    """


# -----------------------------------------------------------------------------


class XarrayYamlDC(XarrayDC):
    """An :py:class:`.XarrayDC` specialization that assumes that each array
    entry is a YAML string, which is subsequently loaded. This can be done
    alongside the metadata application of the :py:class:`.XarrayDC`.
    """

    def _apply_metadata(self):
        """Whenever metadata is applied that is also a good point to resolve
        the YAML data...
        """
        super()._apply_metadata()

        def convert_to_yaml(element) -> dict:
            """Given an array element, try to convert it to yaml"""
            try:
                if isinstance(element, bytes):
                    return _yaml.load(element.decode("utf8"))
                return _yaml.load(element)

            except Exception as exc:
                raise ValueError(
                    f"Could not convert element of type {type(element)} to "
                    f"yaml! Element value was:  {element}"
                ) from exc

        self._data = xr.apply_ufunc(np.vectorize(convert_to_yaml), self._data)


# -----------------------------------------------------------------------------


class GridDC(XarrayDC):
    """This is the base class for all grid data used in Utopia.

    It is based on the :py:class:`.XarrayDC` and reshapes the data to the grid
    shape. The last dimension is assumed to be the dimension that goes along
    the grid cell IDs.
    """

    # Define class variables to allow specializing behaviour ..................
    _GDC_grid_shape_attr = "grid_shape"
    """The attribute to read the desired grid shape from"""

    _GDC_space_extent_attr = "space_extent"
    """The attribute to read the space extent from"""

    _GDC_index_order_attr = "index_order"
    """The attribute to read the index order from"""

    _GDC_grid_structure_attr = "grid_structure"
    """The attribute to read the desired grid structure from"""

    # .........................................................................

    def __init__(
        self, *, name: str, data: Union[np.ndarray, xr.DataArray], **dc_kwargs
    ):
        """Initialize a GridDC which represents grid-like data.

        Given the container attribute (see :py:attr:`._GDC_grid_shape_attr`),
        this container takes care to reshape the underlying data such that it
        represents that grid, even if it is saved in another shape.

        .. note::

            Use this container if it is easier to store array data in a flat
            format (e.g. because there is no need to take care of slicing etc)
            but you still desire to work with it in its actual shape.

        Args:
            name (str): The name of this container
            data (Union[numpy.ndarray, xarray.DataArray]): The not yet
                reshaped data. If this is 1D, it is assumed that there is no
                additional dimension.
                If it is 2D (or more), it is assumed to be ``(..., cell ids)``.
            **kwargs: Further initialization kwargs, e.g. ``attrs`` ...
        """
        # To prohibit proxy resolution in the below __init__, need a flag
        self._shapes_cached = False

        # Call the __init__ function of the base class
        super().__init__(name=name, data=data, **dc_kwargs)
        # ._data is an xr.DataArray now or might still be a proxy

        # Store the old and determine the new shape of the data; needed for
        # proxy property support
        self._data_shape = self.shape

        # Now, determine the shapes ... Will set the following attributes:
        self._new_shape = None
        self._grid_shape = None
        self._new_dims = None
        self._new_coords = None
        self._determine_shape()

        # Have all shapes cached now; set the flag such that the properties
        # use those values instead of the proxy's
        self._shapes_cached = True

        # Do the reshaping and everything only if it is not a proxy
        if not self.data_is_proxy:
            self._data = self._reshape_data()
        # otherwise: postpone reshaping until proxy gets resolved

    def _postprocess_proxy_resolution(self):
        """Invoked from
        :py:class:`~dantro.mixins.proxy_support.Hdf5ProxySupportMixin` after a
        proxy was resolved, this takes care to apply the reshaping operation
        onto the underlying data.
        """
        super()._postprocess_proxy_resolution()
        self._data = self._reshape_data()

    def _parse_sizes_from_metadata(self) -> Sequence[Tuple[str, int]]:
        """Invoked from _format_shape when no metadata was applied but the
        dimension names are available. Should return data in the same form as
        ``xr.DataArray.sizes.items()`` does.
        """
        if not self._shapes_cached:
            return super()._parse_sizes_from_metadata()

        # Iterate over new dimension names and grid shape and use the dim names
        # and shape determined and cached by _determine_shape.
        return tuple(
            (n, l) for i, (n, l) in enumerate(zip(self._new_dims, self.shape))
        )

    # Properties ..............................................................

    @property
    def grid_shape(self) -> tuple:
        """The shape of the grid"""
        if self._shapes_cached:
            return self.shape[1:]
        return self.attrs.get(self._GDC_grid_shape_attr)

    @property
    def space_extent(self) -> tuple:
        """The space's extent this grid is representing, read from attrs"""
        return self.attrs.get(self._GDC_space_extent_attr)

    @property
    def shape(self) -> tuple:
        """Returns shape, proxy-aware

        This is an overload of the property in Hdf5ProxySupportMixin which
        takes care that not the actual underlying proxy data shape is returned
        but whatever the container's shape is to be after reshaping.
        """
        if self.data_is_proxy:
            # Might not be set yet, i.e. during call to super().__init__
            if self._shapes_cached:
                return self._new_shape
            return self.proxy.shape
        return self.data.shape

    @property
    def ndim(self) -> int:
        """Returns ndim, proxy-aware

        This is an overload of the property in Hdf5ProxySupportMixin which
        takes care that not the actual underlying proxy data ndim is returned
        but whatever the container's ndim is to be after reshaping.
        """
        if self.data_is_proxy:
            return len(self.shape)
        return self.data.ndim

    # Reshaping ...............................................................

    def _determine_shape(self):
        """Determine the new shape and store it as ``_new_shape`` attribute"""
        try:
            self._grid_shape = tuple(self.attrs[self._GDC_grid_shape_attr])

        except KeyError as err:
            raise ValueError(
                "Missing attribute '{}' in {} to extract the "
                "desired grid shape from! Available: {}"
                "".format(
                    self._GDC_grid_shape_attr,
                    self.logstr,
                    ", ".join(self.attrs.keys()),
                )
            ) from err

        # To get the new shape, add up the old data shape without the grid
        # dimension and the expected grid shape.
        # NOTE It is assumed that the _last_ dimension goes along cell IDs and
        #      the reshaped x-y data dimensions are thus appended.
        self._new_shape = tuple(self._data_shape[:-1] + self._grid_shape)

        # New shape is now calculated, data shape is still the same
        data_shape = self._data_shape
        new_shape = self._new_shape
        grid_shape = self._grid_shape

        # Determine new dimension names
        if len(data_shape) == 2:
            new_dims = ("time",) + ("x", "y", "z")[: len(grid_shape)]

        elif len(data_shape) == 1:
            new_dims = ("x", "y", "z")[: len(grid_shape)]

        else:
            raise ValueError(
                "Can only reshape from 1D or 2D data, got {}!"
                "".format(data_shape)
            )

        # Determine new coordinates
        new_coords = dict()

        # If there is a space extent attribute available, use that to set the
        # spatial coordinates. Otherwise, assign the trivial coordinates.
        extent = self.attrs.get(self._GDC_space_extent_attr)
        structure = self.attrs.get(self._GDC_grid_structure_attr)
        if extent is None:
            # Trivial coordinate generator
            coord_gen = lambda n, _: range(n)
            extent = (None,) * len(grid_shape)  # dummy for the iterator below
            for n, l, dim_name in zip(grid_shape, extent, ("x", "y", "z")):
                new_coords[dim_name] = coord_gen(n, l)

        elif structure is None or structure == "square":
            # Actual cell position coordinate generator
            coord_gen = lambda n, l: np.linspace(0.0, l, n, False) + (
                0.5 * l / n
            )
            for n, l, dim_name in zip(grid_shape, extent, ("x", "y", "z")):
                new_coords[dim_name] = coord_gen(n, l)

        elif structure == "hexagonal":
            if len(grid_shape) != 2:
                raise ValueError(
                    "Unknown grid structure '{}' in {} dimensions",
                    structure,
                    len(grid_shape),
                )
            new_coords["x"] = (
                np.linspace(0.0, extent[0], grid_shape[0], False)
                + 0.5 * extent[0] / grid_shape[0]
            )
            new_coords["y"] = (
                np.linspace(0.0, extent[1], grid_shape[1], False)
                + 0.5 * extent[1] / grid_shape[1] / 0.75
            )

        else:
            raise ValueError("Unknown grid structure '{}'", structure)

        # NOTE Time coordinates are not changed, thus need not be determined

        # Store the new dimension names and coordinates
        self._new_dims = new_dims
        self._new_coords = new_coords

    def _reshape_data(self) -> xr.DataArray:
        """Looks at the current data shape and container attributes to
        reshape the data such that it represents a grid.
        """
        data_shape = self._data_shape
        new_shape = self._new_shape
        grid_shape = self._grid_shape
        new_dims = self._new_dims
        new_coords = self._new_coords

        # Determine index order
        index_order = self.attrs.get(self._GDC_index_order_attr, "F")

        # Have to postprocess this if order is stored as array of strings
        if isinstance(index_order, np.ndarray):
            if index_order.ndim == 0:
                index_order = index_order.item()
            else:
                index_order = index_order.flatten()[0]

        # Reshape data now
        log.debug(
            "Reshaping data of shape %s to %s (assuming order '%s') to "
            "match given grid shape %s ...",
            data_shape,
            new_shape,
            index_order,
            grid_shape,
        )

        try:
            data = np.reshape(self._data.values, new_shape, order=index_order)

        except ValueError as err:
            raise ValueError(
                "Reshaping failed! This is probably due to a "
                "mismatch between the written dataset attribute "
                "for the grid shape ('{}': {}, configured by "
                "class variable `_GDC_grid_shape_attr`) and the "
                "actual shape {} of the written data."
                "".format(self._GDC_grid_shape_attr, grid_shape, data_shape)
            ) from err

        # All succeeded. Based on the existing data, create a new DataArray
        new_data = xr.DataArray(
            name=self.name,
            data=data,
            dims=new_dims,
            coords=new_coords,
            attrs={
                k: v
                for k, v in self.attrs.items()
                if k
                in (
                    self._GDC_space_extent_attr,
                    self._GDC_grid_shape_attr,
                    self._GDC_grid_structure_attr,
                )
            },
        )

        # Carry over the time coordinates
        if "time" in self._data.dims:
            new_data.coords["time"] = self._data.coords["time"]

        # Done.
        log.debug("Successfully reshaped data to represent a spatial grid.")
        return new_data



"""Implements a single entry of the YAML-based registry framework"""

import copy
import logging
import os
from typing import Any

import pydantic

from .._yaml import load_yml as _load_yml
from .._yaml import write_yml as _write_yml
from ..exceptions import MissingRegistryError, SchemaValidationError

log = logging.getLogger(__name__)


# -----------------------------------------------------------------------------


class BaseSchema(pydantic.BaseModel):
    """A base schema for registry entries.

    This base schema is configured such that it provides safe defaults: Extra
    keys are not allowed and all default values as well as assignments are
    validated.

    .. note::

        The ``validate_assignment`` pydantic configuration option may also lead
        to type coercion! For instance, an integer type value is always
        representable as a string; thus, assigning an integer to a field of
        type string will simply lead to ``str(my_int)`` being stored.

    .. warning::

        If foregoing validation on item assignment, it is no longer guaranteed
        that the written data can be loaded without errors.
    """

    class Config:
        extra = "forbid"
        validate_all = True
        validate_assignment = True

    def __getitem__(self, name: str):
        """Retrieves an item from the underlying schema data."""
        return getattr(self, name)

    def get(self, *args) -> Any:
        """Get a named attribute from this object.

        Behaves exactly like ``getattr(self, *args)``.
        """
        return getattr(self, *args)


class RegistryEntry:
    """A registry entry holds some data (in the form of a validated pydantic
    schema) and allows to read that data from a registry file and write any
    changes to that file."""

    SCHEMA = BaseSchema
    """The data schema that is used for validation"""

    FILE_EXTENSION = ".yml"
    """The file extension that is used for the yaml files -- case-sensitive
    and *with* leading dot.
    """

    _NO_FORWARDING_ATTRS = (
        "_name",
        "_data",
        "_registry",
    )
    """Attribute names that are not forwarded to the underlying data object"""

    def __init__(self, name: str, *, registry: "YAMLRegistry" = None, **data):
        """Initialize a registry entry with a certain name.

        Args:
            name (str): The name of the registry entry, corresponding to a file
                in the registry directory.
            registry (YAMLRegistry, optional): A registry object as part of
                which this entry is managed. If not given, requires ``data``.
            **data: If given, uses this data to initialize the entry. If a
                registry is associated with this entry, will also write that
                data to the corresponding file immediately.
        """
        self._name = name
        self._data = None

        self._registry = None
        self._set_registry(registry)

        # Populate data either from the registry file or the given dict
        if not data:
            if not self.has_registry:
                raise MissingRegistryError(
                    f"To construct a {type(self).__name__} without data, a "
                    "registry needs to be associated with it!"
                )

            self.load()
            log.debug("%s initialized from file.", self)

        else:
            self._data = self._parse_data(data)
            if self.has_registry:
                self.write()
                log.debug(
                    "%s initialized from data and written to file.", self
                )
            else:
                log.debug(
                    "%s initialized from data without writing to file.", self
                )

    def _parse_data(self, d: dict) -> BaseSchema:
        """Uses the schema to set the entry's data from the given dict"""
        try:
            return self.SCHEMA(**d)

        except pydantic.ValidationError as err:
            raise SchemaValidationError(
                f"Failed parsing data into {self}!\n{err}"
            ) from err

    def _set_registry(self, registry: "YAMLRegistry"):
        """Associates a registry with this entry"""
        self._registry = registry

    # Properties and data access ..............................................

    @property
    def name(self) -> str:
        """Name of this entry"""
        return self._name

    @property
    def has_registry(self) -> bool:
        """Whether a registry is associated with this entry"""
        return self._registry is not None

    @property
    def registry_dir(self) -> str:
        """The associated registry directory"""
        return self._registry.registry_dir

    @property
    def registry_file_path(self) -> str:
        """The absolute path to the registry file"""
        return os.path.join(
            self.registry_dir, f"{self.name}{self.FILE_EXTENSION}"
        )

    @property
    def data(self) -> BaseSchema:
        """The entry's data"""
        return self._data

    def dict(self) -> dict:
        """The entry's data in pydantic's dict format, deep-copied."""
        return copy.deepcopy(self._data.dict())

    # Magic methods ...........................................................

    def __str__(self) -> str:
        """String descriptor of this object"""
        return f"<{type(self).__name__} '{self.name}'>"

    def __repr__(self) -> str:
        return f"<{type(self).__name__} '{self.name}': {self._data}>"

    def __eq__(self, other: Any) -> bool:
        """An entry compares equal if the type is identical and the name and
        data compare equal.

        .. note::

            The associated registry is not compared!
        """
        if type(self) is not type(other):
            return False
        return self._name == other._name and self._data == other._data

    def __getattr__(self, attr: str):
        """Forward attribute calls (that do not match the entry's other
        attributes) to the underlying entry data.

        Alternatively, use the ``data`` property to directly access the data.
        """
        return getattr(self._data, attr)

    def __setattr__(self, attr: str, value: Any):
        """Forwards attribute setting calls to the underlying entry data.

        This is only done if ``attr`` is not in ``_NO_FORWARDING_ATTRS`` and
        the ``attr`` is actually a property of the schema. Otherwise, regular
        attribute setting occurs.

        .. warning::

            Changes to the data are *not* automatically written to the YAML
            file. To do so, call
            :py:meth:`~utopya._yaml_registry.entry.RegistryEntry.write` after
            all changes have been completed.

            Note that validation will occur at that point and not when changing
            any data values.
        """
        if (
            attr in type(self)._NO_FORWARDING_ATTRS
            or attr not in self._data.schema()["properties"]
        ):
            return super().__setattr__(attr, value)
        return setattr(self._data, attr, value)

    def __getitem__(self, name: str):
        """Retrieves an item from the underlying entry data."""
        return getattr(self._data, name)

    def get(self, *args) -> Any:
        """Get a named attribute from this object's entry or return a fallback.

        Behaves exactly like ``getattr(self, *args)``.
        """
        return getattr(self._data, *args)

    # Loading and writing data ................................................

    def load(self):
        """Reads the entry from the registry file, raising an error if the
        file does not exist. The loaded data overwrites whatever is stored in
        the entry already.
        """
        if not self.has_registry:
            raise MissingRegistryError(
                f"No registry associated with {self}, cannot load entry data!"
            )

        if not os.path.isfile(self.registry_file_path):
            raise FileNotFoundError(
                f"Missing registry file for {self} "
                f"at {self.registry_file_path}!"
            )

        try:
            d = _load_yml(self.registry_file_path)

        except Exception as exc:
            raise type(exc)(
                f"Failed loading registry file {self.registry_file_path}! "
                "See traceback for more information."
            ) from exc

        self._data = self._parse_data(d)

    def write(self):
        """Writes the registry entry to the corresponding registry file,
        creating it if it does not exist.

        .. note::

            The data is written with an intermediate json-serialization
            carried out by pydantic. That ensures that the data contains only
            native data types which can be written to YAML without any custom
            representers.
        """
        if not self.has_registry:
            raise MissingRegistryError(
                f"No registry associated with {self}, cannot write entry data!"
            )

        import json

        data = json.loads(self.data.json())
        _write_yml(data, path=self.registry_file_path)

    def remove_registry_file(self):
        """Removes the corresponding registry file"""
        if not self.has_registry:
            raise MissingRegistryError(
                f"No registry associated with {self}, cannot remove file!"
            )

        os.remove(self.registry_file_path)



"""Implements a plotting framework based on dantro

In order to make the plotting framework specific to Utopia, this module derives
both from the dantro PlotManager and some PlotCreator classes.
"""

import copy
import importlib
import logging
import os
import sys
from typing import Union

import dantro
import dantro.plot.creators
import dantro.plot.utils
import dantro.plot_mngr

from ..model_registry import ModelInfoBundle
from ._plot_func_resolver import PlotFuncResolver
from .plotcreators import (
    MultiversePlotCreator,
    PyPlotCreator,
    UniversePlotCreator,
)

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


class PlotManager(dantro.plot_mngr.PlotManager):
    """This is the Utopia-specific version of the dantro ``PlotManager``.

    It registers the Utopia-specific plot creators and allows for custom
    interface specifications, e.g. by preloading custom modules.
    """

    CREATORS = dict(
        base=dantro.plot.creators.BasePlotCreator,
        external=PyPlotCreator,
        pyplot=PyPlotCreator,
        universe=UniversePlotCreator,
        multiverse=MultiversePlotCreator,
    )
    """Supported plot creator classes"""

    MODEL_PLOTS_MODULE_NAME = "model_plots"
    """Name under which the model-specific plots are made importable"""

    PLOT_FUNC_RESOLVER: type = PlotFuncResolver
    """The custom plot function resolver type to use."""

    # .........................................................................

    def __init__(
        self, *args, _model_info_bundle: ModelInfoBundle = None, **kwargs
    ):
        """Sets up a PlotManager.

        This additionally stores some Utopia-specific metadata about the
        model this PlotManager is used with. That information is then used to
        load some additional model-specific information once a creator is
        invoked.
        """
        super().__init__(*args, **kwargs)

        self._model_info_bundle = copy.deepcopy(_model_info_bundle)

    @property
    def common_out_dir(self) -> str:
        """The common output directory of all plots that were created with
        this plot manager instance. This uses the plot output paths stored in
        the plot information dict, specifically the ``target_dir`` entry.

        If there was no plot information yet, the return value will be empty.
        """
        p = os.path.commonprefix([d["target_dir"] for d in self.plot_info])
        if not os.path.exists(p):
            p = os.path.dirname(p)
        return p

    def plot_from_cfg(
        self, *args, plots_cfg: Union[str, dict] = None, **kwargs
    ):
        """Thin wrapper around parent method that shows which plot
        configuration file will be used.
        """
        log.hilight(
            "Now creating plots for '%s' model ...",
            self._model_info_bundle.model_name,
        )
        if isinstance(plots_cfg, str):
            log.note("Plots configuration:\n  %s\n", plots_cfg)

        elif plots_cfg is None:
            log.note(
                "Using default plots configuration:\n  %s\n",
                self._model_info_bundle.paths.get("default_plots"),
            )

        return super().plot_from_cfg(*args, plots_cfg=plots_cfg, **kwargs)

    def _get_plot_func_resolver(self, **init_kwargs) -> PlotFuncResolver:
        """Instantiates the plot function resolver object.

        Additionally attaches the model info bundle to the resolver, such that
        it can use that information for plot function lookup.
        """
        return self.PLOT_FUNC_RESOLVER(
            **init_kwargs, _model_info_bundle=self._model_info_bundle
        )

    def _get_plot_creator(self, *args, **kwargs):
        """Before actually retrieving the plot creator, invokes module
        pre-loading via :py:meth:`._preload_modules`.
        """
        self._preload_modules()

        creator = super()._get_plot_creator(*args, **kwargs)
        creator._model_info_bundle = copy.deepcopy(self._model_info_bundle)

        return creator

    def _preload_modules(self):
        """Pre-loads the model-, project-, and framework-specific plot
        function modules.
        This allows to execute code (like registering model-specific dantro
        data operations) and have them available prior to the invocation of
        the creator and independently from the module that contains the plot
        function (which may be part of dantro, for instance).
        """

        from .._import_tools import import_module_from_path

        mib = self._model_info_bundle

        if mib is not None:
            log.note("Pre-loading plot modules ...")
            mod_path = mib.paths.get("py_plots_dir")
            if mod_path:
                log.remark("  Loading model-specific modules ...")
                import_module_from_path(
                    mod_path=mod_path,
                    mod_str=f"{self.MODEL_PLOTS_MODULE_NAME}.{mib.model_name}",
                    debug=self.raise_exc,
                )

            # Also do this on the project and framework level
            # TODO Should make module name configurable separately! See #9
            project = mib.project
            if project and project.paths.py_plots_dir:
                log.remark("  Loading project-specific modules ...")
                import_module_from_path(
                    mod_path=project.paths.py_plots_dir,
                    mod_str=f"{self.MODEL_PLOTS_MODULE_NAME}",
                    debug=self.raise_exc,
                )

            if project and project.framework_project:
                fw = project.framework_project
                if fw and fw.paths.py_plots_dir:
                    log.remark("  Loading framework-specific modules ...")
                    import_module_from_path(
                        mod_path=fw.paths.py_plots_dir,
                        mod_str=f"{self.MODEL_PLOTS_MODULE_NAME}",
                        debug=self.raise_exc,
                    )
